"""
The Proxy module provides the base class for the Proxy objects for each device
controller.

The module also provides the connection state interface and classes for
maintaining the state of the Proxy connection to the control server.
"""

import logging
import pickle
import types
from types import MethodType

import zmq

from egse.connect import VERBOSE_DEBUG
from egse.decorators import dynamic_interface
from egse.log import logger
from egse.mixin import DynamicClientCommandMixin
from egse.response import Failure
from egse.zmq_ser import split_address


def set_docstring(func, cmd):
    """Decorator to set the docstring of the command on the dynamic method / function."""

    def wrap_func(*args, **kwargs):
        return func(*args, **kwargs)

    wrap_func.__doc__ = cmd.__doc__
    wrap_func.__name__ = f"{cmd.get_name()}"
    return wrap_func


REQUEST_TIMEOUT = 30.0  # timeout in seconds
REQUEST_RETRIES = 0


class ControlServerConnectionInterface:
    """This interface defines the connection commands for control servers.

    This interface shall be implemented by the Proxy class and guarantees that connection commands
    do not interfere with the commands defined in the `DeviceConnectionInterface` (which will be
    loaded from the control server).
    """

    def __enter__(self):
        self.connect_cs()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect_cs()

    @dynamic_interface
    def connect_cs(self):
        """Connect to the control server.

        Raises:
            ConnectionError: when the connection can not be established.
        """
        raise NotImplementedError

    @dynamic_interface
    def reconnect_cs(self):
        """Reconnect the control server after it has been disconnected.

        Raises:
            ConnectionError: when the connection can not be established.
        """
        raise NotImplementedError

    @dynamic_interface
    def disconnect_cs(self):
        """Disconnect from the control server.

        Raises:
            ConnectionError: when the connection can not be closed.
        """
        raise NotImplementedError

    @dynamic_interface
    def reset_cs_connection(self):
        """Resets the connection to the control server."""
        raise NotImplementedError

    @dynamic_interface
    def is_cs_connected(self) -> bool:
        """Check if the control server is connected.

        Returns:
            True if the device is connected and responds to a command, False otherwise.
        """
        raise NotImplementedError


class BaseProxy(ControlServerConnectionInterface):
    def __init__(self, endpoint: str, timeout: float = REQUEST_TIMEOUT):
        """
        The endpoint is a string that is constructed from the protocol, hostname
         and port number and has the format: `protocol://hostname:port`.

        The `timeout` argument specifies the number of fractional seconds to wait
        for a reply from the control server.
        """

        self._logger = logger

        self._ctx = zmq.Context.instance()
        self._poller = zmq.Poller()
        self._socket = None
        self._endpoint = endpoint
        self._timeout = timeout

        self.connect_cs()

    def __enter__(self):
        if not self.ping():
            raise ConnectionError(f"Proxy is not connected to endpoint ({self._endpoint}) when entering the context.")

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._socket.closed:
            self.disconnect_cs()

    def connect_cs(self):
        if VERBOSE_DEBUG:
            self._logger.debug(f"Trying to connect {self.__class__.__name__} to {self._endpoint}")

        self._socket = self._ctx.socket(zmq.REQ)
        self._socket.connect(self._endpoint)
        self._poller.register(self._socket, zmq.POLLIN)

    def disconnect_cs(self):
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.close()
        self._poller.unregister(self._socket)

    def reconnect_cs(self):
        if VERBOSE_DEBUG:
            self._logger.debug(f"Trying to reconnect {self.__class__.__name__} to {self._endpoint}")

        if not self._socket.closed:
            self._socket.close(linger=0)

        self._socket = self._ctx.socket(zmq.REQ)
        self._socket.connect(self._endpoint)
        self._poller.register(self._socket, zmq.POLLIN)

    def reset_cs_connection(self):
        if VERBOSE_DEBUG:
            self._logger.debug(f"Trying to reset the connection from {self.__class__.__name__} to {self._endpoint}")

        self.disconnect_cs()
        self.connect_cs()

    def is_cs_connected(self) -> bool:
        return self.ping()

    def send(self, data, retries: int = REQUEST_RETRIES, timeout: float | None = None):
        """
        Sends a command to the control server and waits for a response.

        When not connected to the control server or when a timeout occurs, the
        ``send()`` command retries a number of times to send the command.

        The number of retries is hardcoded and currently set to '2', the request
        timeout is set to 2.5 seconds.

        The command data will be pickled before sending. Make sure the ``data``
        argument can be dumped by pickle.

        Args:
            data (str): the command that is sent to the control server, usually a
                string, but that is not enforced.
            timeout (int): the time to wait for a reply [in seconds]
            retries (int): the number of time we should retry to send the message

        Returns:
            response: the response from the control server or ``None`` when there was
                a problem or a timeout.
        """
        timeout_ms = int((timeout or self._timeout) * 1000)

        pickle_string = pickle.dumps(data)

        retries_left = retries

        # When we enter this method, we assume the Proxy has been connected. It
        # might be the server is not responding, but that is handled by the
        # algorithm below where we have a number of retries to receive the response
        # of the sent command. Remember that we are using ZeroMQ where the connect
        # method returns gracefully even when no server is available.

        if self._socket.closed:
            self.reconnect_cs()

        if VERBOSE_DEBUG:
            self._logger.debug(f"Sending '{data}'")

        self._socket.send(pickle_string)

        while True:
            socks = dict(self._poller.poll(timeout_ms))

            if self._socket in socks and socks[self._socket] == zmq.POLLIN:
                pickle_string = self._socket.recv()
                if not pickle_string:
                    break
                response = pickle.loads(pickle_string)
                if VERBOSE_DEBUG:
                    self._logger.debug(f"Receiving response: {response}")
                return response
            else:
                # timeout - server unavailable

                # We should disconnect here because socket is possibly confused.
                # Close the socket and remove from the poller.

                self.disconnect_cs()

                if retries_left == 0:
                    self._logger.critical(f"Control Server seems to be off-line, abandoning ({data})")
                    return Failure(f"Control Server seems to be off-line, abandoning ({data})")
                retries_left -= 1

                self._logger.log(logging.CRITICAL, f"Reconnecting {self.__class__.__name__}, {retries_left=}")

                self.reconnect_cs()

                # Now try to send the request again

                self._socket.send(pickle_string)

    def ping(self):
        return_code = self.send("Ping", retries=0, timeout=1.0)
        if VERBOSE_DEBUG:
            self._logger.debug(f"Check if control server is available: Ping - {return_code}")
        return return_code == "Pong"

    def get_endpoint(self) -> str:
        """Returns the endpoint."""
        return self._endpoint

    def get_monitoring_port(self) -> int:
        """Returns the monitoring port."""
        return self.send("get_monitoring_port")

    def get_commanding_port(self) -> int:
        """Returns the commanding port."""
        return self.send("get_commanding_port")

    def get_service_port(self) -> int:
        """Returns the service port."""
        return self.send("get_service_port")

    def get_ip_address(self) -> int:
        """Returns the hostname of the control server."""
        return self.send("get_ip_address")

    def get_service_proxy(self):
        """Return a ServiceProxy for the control server of this proxy object."""
        from egse.services import ServiceProxy  # prevent circular import problem

        transport, address, _ = split_address(self._endpoint)

        response = self.send("get_service_port")  # FIXME: Check if this is still returning the proper port

        logger.debug(f"----> {response=}")

        if isinstance(response, Failure):
            raise response
        else:
            return ServiceProxy(protocol=transport, hostname=address, port=response)


class DynamicProxy(BaseProxy, DynamicClientCommandMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# TODO (rik): remove all methods from Proxy that are also define in the BaseProxy


class Proxy(BaseProxy, ControlServerConnectionInterface):
    """
    A Proxy object will forward CommandExecutions to the connected control server
    and wait for a response. When the Proxy can not connect to its control server
    during initialization, a ConnectionError will be raised.
    """

    def __init__(self, endpoint, timeout: float = REQUEST_TIMEOUT):
        """
        During initialization, the Proxy will connect to the control server and send a
        handshaking `Ping` command. When that succeeds the Proxy will request and load the
        available commands from the control server. When the connection with the control server
        fails, no commands are loaded and the Proxy is left in a 'disconnected' state. The caller
        can fix the problem with the control server and call `connect_cs()`, followed by a call to
        `load_commands()`.

        The `timeout` argument specifies the number of seconds
        """

        super().__init__(endpoint, timeout)

        self._commands = {}

        if self.ping():
            self.load_commands()
        else:
            self._logger.warning(
                f"{self.__class__.__name__} could not connect to its control server at {endpoint}. "
                f"No commands have been loaded."
            )

    def __enter__(self):
        if not self.ping():
            raise ConnectionError("Proxy is not connected when entering the context.")

        # The following check is here because a CS might have come alive between the __init__
        # and __enter__ calls, and while the ping() will reconnect, the Proxy will have no
        # commands loaded.

        if not self.has_commands():
            self.load_commands()

        return self

    def _request_commands(self):
        response = self.send("send_commands")
        if isinstance(response, Failure):
            raise response
        self._commands = response

    def _add_commands(self):
        for key in self._commands:
            if hasattr(self, key):
                attribute = getattr(self, key)
                if isinstance(attribute, types.MethodType) and not hasattr(attribute, "__dynamic_interface"):
                    self._logger.warning(
                        f"{self.__class__.__name__} already has an attribute '{key}', not overwriting."
                    )
                    continue
            command = self._commands[key]
            new_method = MethodType(command.client_call, self)
            new_method = set_docstring(new_method, command)
            setattr(self, key, new_method)

    def load_commands(self):
        """
        Requests all available commands from the control server and adds them to
        the Proxy public interface, i.e. each command will become a method for
        this Proxy.

        A warning will be issued when a command will overwrite an existing method
        of the Proxy class. The original method will not be overwritten and the
        behavior of the Proxy command will not be what is expected.
        """
        # bind the client_call method from each Command to this Proxy object
        if self.is_cs_connected():
            try:
                self._request_commands()
            except Failure as exc:
                self._logger.warning(f"Failed to request commands for {self.__class__.__name__}: {exc}")
                return False
            else:
                self._add_commands()
                return True
        else:
            self._logger.warning(f"{self.__class__.__name__} is not connected, try to reconnect.")
            return False

    def get_commands(self):
        """
        Returns a list of command names that can be send to the device or the
        control server.

        The commands are defined in the YAML settings file of the device.
        Special commands are available for the ServiceProxy which configure and
        control the control servers.
        """
        return list(self._commands.keys())

    def has_commands(self):
        """Return `True` if commands have been loaded."""
        return bool(self._commands)



"""
Unit tests for the observation module building blocks and BBID generation.
"""

from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from egse.observation import building_block
from egse.observation import end_building_block
from egse.observation import get_bbid_uuid
from egse.observation import start_building_block


def test_building_block_sets_attribute():
    """Test that the decorator sets the __building_block_func attribute."""

    @building_block
    def my_func(x: int) -> int:
        return x * 2

    assert hasattr(my_func, "__building_block_func")
    assert getattr(my_func, "__building_block_func") is True


def test_building_block_preserves_function_name():
    """Test that the decorator preserves the original function name."""

    @building_block
    def my_test_func(x: int) -> int:
        return x

    assert my_test_func.__name__ == "my_test_func"


@patch("egse.observation.ObservationContext")
def test_building_block_executes_function(mock_obs_context):
    """Test that the decorator properly executes the wrapped function."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def add(a: int, b: int) -> int:
        return a + b

    result = add(a=2, b=3)

    assert result == 5

    # Verify that ObservationContext methods were called
    mock_instance.set_bbid.assert_called_once()
    mock_instance.unset_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_building_block_rejects_positional_arguments(mock_obs_context):
    """Test that the decorator raises ValueError for positional arguments."""

    @building_block
    def my_func(x: int) -> int:
        return x

    with pytest.raises(ValueError, match="can not have positional arguments"):
        my_func(5)


@patch("egse.observation.ObservationContext")
def test_building_block_rejects_missing_kwargs(mock_obs_context):
    """Test that the decorator raises ValueError for missing keyword arguments."""

    @building_block
    def my_func(x: int, y: int) -> int:
        return x + y

    with pytest.raises(ValueError, match="missing arguments"):
        my_func(x=1)  # Missing y


@patch("egse.observation.ObservationContext")
def test_building_block_accepts_all_kwargs(mock_obs_context):
    """Test that the decorator accepts functions with all required kwargs."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def my_func(x: int, y: int) -> int:
        return x + y

    result = my_func(x=1, y=2)

    assert result == 3


@patch("egse.observation.ObservationContext")
def test_building_block_calls_cleanup_on_exception(mock_obs_context):
    """Test that unset_bbid is called even if the function raises an exception."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def failing_func():
        raise RuntimeError("test error")

    with pytest.raises(RuntimeError, match="test error"):
        failing_func()

    # Verify that unset_bbid was still called
    mock_instance.unset_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_building_block_with_no_parameters(mock_obs_context):
    """Test that the decorator works with functions that have no parameters."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def no_params_func() -> int:
        return 42

    result = no_params_func()

    assert result == 42
    mock_instance.set_bbid.assert_called_once()
    mock_instance.unset_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_building_block_with_lambda(mock_obs_context):
    """Test that the decorator works with lambda functions."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    decorated_lambda = building_block(lambda x: x * 2)

    result = decorated_lambda(x=5)

    assert result == 10


@patch("egse.observation.ObservationContext")
def test_building_block_function_returning_none(mock_obs_context):
    """Test that building blocks that return None are handled correctly."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def returns_none():
        return None

    result = returns_none()

    assert result is None
    mock_instance.set_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_building_block_with_multiple_parameters(mock_obs_context):
    """Test building block with various parameter types."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def complex_func(a: int, b: str, c: float) -> str:
        return f"{b}: {a + c}"

    result = complex_func(a=1, b="result", c=2.5)

    assert result == "result: 3.5"
    mock_instance.set_bbid.assert_called_once()
    mock_instance.unset_bbid.assert_called_once()


def test_get_bbid_uuid_uppercase_function_name():
    """Test that the function name is converted to uppercase."""

    @building_block
    def my_test_function():
        pass

    name = get_bbid_uuid(my_test_function)

    assert "MY_TEST_FUNCTION" in name


def test_generate_bbid_correct_prefix():
    """Test that generated BBID starts with BBID prefix."""

    @building_block
    def my_func():
        pass

    bbid = get_bbid_uuid(my_func)

    assert bbid.startswith("BBID")


def test_generate_bbid_correct_length():
    """Test that generated BBID has correct length (BBID + 16 hex chars)."""

    @building_block
    def my_func():
        pass

    bbid = get_bbid_uuid(my_func)

    # BBID prefix (4 chars) + '-' + len("my_func") + '-' + 64 hex chars = 77 chars total
    assert len(bbid) == 77


def test_generate_bbid_is_hex_postfix():
    """Test that the BBID suffix is valid hexadecimal."""

    @building_block
    def my_func():
        pass

    bbid = get_bbid_uuid(my_func)
    hex_part = bbid[13:]  # Remove BBID prefix + '_' + len("my_func") + '_' = 13 chars

    # Should not raise ValueError
    int(hex_part, 16)


def test_generate_bbid_consistent():
    """Test that the same function always generates the same BBID."""

    @building_block
    def test_func():
        pass

    bbid1 = get_bbid_uuid(test_func)
    bbid2 = get_bbid_uuid(test_func)

    assert bbid1 == bbid2


def test_generate_bbid_different_functions():
    """Test that different functions generate different BBIDs."""

    @building_block
    def func1():
        pass

    @building_block
    def func2():
        pass

    bbid1 = get_bbid_uuid(func1)
    bbid2 = get_bbid_uuid(func2)

    assert bbid1 != bbid2


def test_generate_bbid_uppercase():
    """Test that BBID is always uppercase."""

    @building_block
    def my_func():
        pass

    bbid = get_bbid_uuid(my_func)

    assert bbid == bbid.upper()


def test_get_bbid_uuid_raises_error_non_callable():
    """Test that get_bbid_uuid raises ValueError for non-callable."""
    with pytest.raises(ValueError, match="should be a function or method"):
        get_bbid_uuid("not a function")  # type: ignore


def test_get_bbid_uuid_raises_error_none():
    """Test that get_bbid_uuid raises ValueError for None."""
    with pytest.raises(ValueError, match="should be a function or method"):
        get_bbid_uuid(None)  # type: ignore


def test_get_bbid_uuid_raises_error_not_building_block():
    """Test that get_bbid_uuid raises ValueError for functions not defined as building blocks."""

    def regular_function():
        pass

    with pytest.raises(ValueError, match="not defined as a building block"):
        get_bbid_uuid(regular_function)


@patch("egse.observation.ObservationContext")
@patch("egse.observation.get_bbid_uuid")
def test_start_building_block_sets_bbid(mock_get_bbid, mock_obs_context):
    """Test that start_building_block sets the BBID in ObservationContext."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance
    mock_get_bbid.return_value = "BBID1234567890ABCDEF"

    def my_func():
        pass

    start_building_block(my_func)

    mock_instance.set_bbid.assert_called_once_with("BBID1234567890ABCDEF")


@patch("egse.observation.ObservationContext")
@patch("egse.observation.get_bbid_uuid")
def test_start_building_block_returns_bbid(mock_get_bbid, mock_obs_context):
    """Test that start_building_block returns the BBID."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance
    mock_get_bbid.return_value = "BBID_TEST_VALUE"

    def my_func():
        pass

    result = start_building_block(my_func)

    assert result == "BBID_TEST_VALUE"


@patch("egse.observation.ObservationContext")
def test_end_building_block_unsets_bbid(mock_obs_context):
    """Test that end_building_block unsets the BBID in ObservationContext."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    end_building_block()

    mock_instance.unset_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_full_building_block_lifecycle(mock_obs_context):
    """Test the complete lifecycle of a building block execution."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def process_data(value: int) -> int:
        return value * 2

    result = process_data(value=5)

    # Verify function executed correctly
    assert result == 10

    # Verify ObservationContext was properly managed
    mock_instance.set_bbid.assert_called_once()
    mock_instance.unset_bbid.assert_called_once()


@patch("egse.observation.ObservationContext")
def test_multiple_building_blocks(mock_obs_context):
    """Test using multiple decorated functions."""
    mock_instance = MagicMock()
    mock_obs_context.return_value = mock_instance

    @building_block
    def func1(x: int) -> int:
        return x + 1

    @building_block
    def func2(y: int) -> int:
        return y * 2

    result1 = func1(x=5)
    result2 = func2(y=3)

    assert result1 == 6
    assert result2 == 6

    # Each function call should set and unset BBID
    assert mock_instance.set_bbid.call_count == 2
    assert mock_instance.unset_bbid.call_count == 2



from egse.ariel.tcu.tcu_cmd_utils import create_crc16


def test_create_crc16():
    cmd_string = " 0340 0001 0004 0000 0000 0000 0000"
    expected_crc = "3998"

    assert create_crc16(cmd_string) == expected_crc



import subprocess
from typing import Annotated

import rich
import sys
import typer

from egse.system import redirect_output_to_log

tgf4000 = typer.Typer(name="tgf4000", help="Aim -TTi, Dual-channel arbitrary function generator", no_args_is_help=True)


@tgf4000.command(name="start")
def start_tgf40000(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="use a device simulator as the backend")
    ] = False,
):
    """Starts the TGF4000 service.

    Args:
        device_id: TGF4000-A identifier.
    """

    rich.print("Starting service tgf4000")
    out = redirect_output_to_log("tgf4000.start.log")

    cmd = [sys.executable, "-m", "egse.arbitrary_wave_generator.aim_tti.tgf4000_cs", "start", device_id]
    if simulator:
        cmd.append("--simulator")
    subprocess.Popen(
        cmd,
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@tgf4000.command(name="stop")
def stop_tgf4000(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Stops the TGF4000 service.

    Args:
        device_id: TGF4000 identifier.
    """

    rich.print("Terminating service TGF4000")

    out = redirect_output_to_log("tgf4000_cs.stop.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.arbitrary_wave_generator.aim_tti.tgf4000_cs", "stop", device_id],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@tgf4000.command(name="status")
def status_tgf4000(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Prints status information on the TGF4000 service.

    Args:
        device_id: TGF4000 identifier.
    """

    proc = subprocess.Popen(
        [sys.executable, "-m", "egse.arbitrary_wave_generator.aim_tti.tgf4000_cs", "status", device_id],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.DEVNULL,
    )

    stdout, stderr = proc.communicate()

    rich.print(stdout.decode(), end="")


@tgf4000.command(name="start-sim")
def start_tgf4000_sim(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Start the TGF4000 Simulator.

    Args:
        device_id: TGF4000 identifier.
    """

    rich.print("Starting service TGF4000 Simulator")

    out = redirect_output_to_log("tgf4000_sim.start.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.arbitrary_wave_generator.aim_tti.tgf4000_sim", "start", device_id],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


@tgf4000.command(name="stop-sim")
def stop_tgf4000_sim(
    device_id: Annotated[str, typer.Argument(help="the device identifier, identifies the hardware controller")],
):
    """Stops the TGF4000 Simulator.

    Args:
        device_id: TGF4000 identifier.
    """

    rich.print("Terminating the TGF4000 simulator.")

    out = redirect_output_to_log("tgf4000_sim.stop.log")

    subprocess.Popen(
        [sys.executable, "-m", "egse.arbitrary_wave_generator.aim_tti.tgf4000_sim", "stop", device_id],
        stdout=out,
        stderr=out,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )


if __name__ == "__main__":
    tgf4000()



import re

from egse.env import bool_env
from egse.hexapod.symetrie import logger

VERBOSE_DEBUG = bool_env("VERBOSE_DEBUG")

regex_response = {
    # Error message as ERRXXX
    "ERROR": re.compile(r"\a(ERR\d{3})\r"),
    # a floating point number [-]x[.[xxx]], possibly followed by spaces
    "FLOAT": re.compile(r"(-?(\d*\.)?\d+)\s*\r\x06"),
    # an integer number possible [-]iiiii, possibly followed by spaces
    "INT": re.compile(r"(-?\d+)\s*\r\x06"),
    # Anything else, stripping off '\x06'
    "ANY": re.compile(r"(.*)\x06"),
    # Just the null character
    "NUL": re.compile(r"(\x00)"),
}


def match_regex_response(regex_prog, res):
    """
    This 'matches' from the start of the string 'res'. If you want to match anywhere in the
    string, you will need to make another function search_regex_response() with 'search' instead
    of 'match'.

    Return None if no match and the match object otherwise.
    """
    if VERBOSE_DEBUG:
        logger.debug(f"res = {res} with type {type(res)}")
    if isinstance(res, bytes):
        res = res.decode()
    match_obj = regex_prog.match(res)
    return match_obj


def patter_n_int(nr):
    return re.compile(rf"(-?\d+)\r((-?\d+)\r){{{nr}}}\x06")


def patter_n_float(nr):
    return re.compile(rf"(-?\d+)\r((-?(\d*\.)?\d+)\r){{{nr}}}\x06")


def match_int_response(res):
    match_obj = match_regex_response(regex_response["INT"], res)
    if match_obj is None:
        logger.error(f"Could not parse INT response for {res}")
        return None
    return int(match_obj[1])


def match_float_response(res):
    match_obj = match_regex_response(regex_response["FLOAT"], res)
    if match_obj is None:
        logger.error(f"Could not parse FLOAT response for {res}")
        return None
    return float(match_obj[1])


def match_error_response(res):
    match_obj = match_regex_response(regex_response["ERROR"], res)
    if match_obj is None:
        logger.error(f"Could not parse ERROR response for {res}")
        return None
    return match_obj[1]


def match_nul_response(res):
    match_obj = match_regex_response(regex_response["NUL"], res)
    if match_obj is None:
        logger.error(f"Could not parse NUL response for {res}")
        return None
    return match_obj[1]


def match_string_response(res):
    match_obj = match_regex_response(regex_response["ANY"], res)
    if match_obj is None:
        logger.error(f"Could not parse STRING response for {res}")
        return None
    if len(match_obj[1]) > 0 and match_obj[1][-1] == "\r":
        return match_obj[1][:-1]
    else:
        return match_obj[1]



# Unit Tests were created by GitHub Copilot and adapted/improved afterward.

import asyncio
import socketserver
import threading
import time

import pytest
import pytest_asyncio

from egse.device import DeviceConnectionError
from egse.device import DeviceTimeoutError
from egse.log import logging
from egse.socketdevice import AsyncSocketDevice
from egse.socketdevice import SocketDevice
from egse.system import type_name

logger = logging.getLogger("egse.test.socketdevice")

SEPARATOR = b"\x03"
SEPARATOR_STR = SEPARATOR.decode()


# Synchronous server used by SocketDevice tests
class _SyncEchoHandler(socketserver.BaseRequestHandler):
    """
    Echo handler: sends back a prefix and an ETX (b'\x03') so client read() completes.
    Keeps connection open until client closes socket.
    """

    def handle(self):
        try:
            while True:
                data = self.request.recv(4096)
                if not data:
                    break
                if b"NO-REPLY" in data:
                    continue
                if b"REPLY-IN-PARTS" in data:
                    self.request.sendall(b"RESP:")
                    # time.sleep(0.2)
                    self.request.sendall(data)
                    # time.sleep(0.2)
                    self.request.sendall(SEPARATOR)
                else:
                    self.request.sendall(b"RESP:" + data + SEPARATOR)
        except Exception as exc:
            logger.warning(f"Caught {type_name(exc)}: {exc}")


# Properly close the async_echo_server including active writers
async def close_server(server, active_writers):
    # This coroutine is used by the async_echo_server fixture for the teardown of that fixture,
    # and it is also used by the test_async_cleanup_on_broken_connection test to make sure no active
    # connections exist and the test can succeed.

    server.close()

    try:
        await asyncio.wait_for(server.wait_closed(), timeout=1.0)
    except asyncio.TimeoutError:
        logger.warning("server.wait_closed() timed out; continuing to close active writers")

    logger.info("After server close")

    # explicitly close all existing client connections
    for active_writer in list(active_writers):
        try:
            logger.debug(f"Closing {active_writer=}")
            active_writer.close()
            await active_writer.wait_closed()
        except Exception as exc:
            logger.warning(f"Caught {type_name(exc)}: {exc}")

    logger.info("Server and active connections terminated.")


# Synchronous server used by SocketDevice tests
@pytest.fixture(scope="module")
def sync_echo_server():
    """Start a ThreadingTCPServer serving _SyncEchoHandler on localhost:0 (ephemeral dynamic port)."""
    server = socketserver.ThreadingTCPServer(("127.0.0.1", 0), _SyncEchoHandler)
    server.allow_reuse_address = True
    host, port = server.server_address

    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()

    yield host, port

    server.shutdown()
    server.server_close()
    thread.join(timeout=1)


# Asynchronous server used by AsyncSocketDevice tests
@pytest_asyncio.fixture
async def async_echo_server():
    """
    Start an asyncio server that echos back data with a prefix and an ETX (b'\x03').
    Yields (host, port, server, active_writers) then closes the server on teardown.

    The active writers are used in the test to handle a broken connection.
    """
    active_writers = set()

    async def handler(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        active_writers.add(writer)
        try:
            while True:
                data = await reader.read(4096)
                if not data:
                    break
                if b"NO-REPLY" in data:
                    continue
                writer.write(b"RESP:" + data + SEPARATOR)
                await writer.drain()
        except Exception as exc:
            logger.warning(f"Caught {type_name(exc)}: {exc}")
        finally:
            active_writers.discard(writer)
            try:
                writer.close()
                await writer.wait_closed()
            except Exception as exc:
                logger.warning(f"Caught {type_name(exc)}: {exc}")

    server = await asyncio.start_server(handler, "127.0.0.1", 0)
    sockets = server.sockets

    logger.debug(f"{sockets=}, {sockets[0].getsockname()=}")

    assert sockets, "Server sockets missing"
    host, port = sockets[0].getsockname()[:2]  # getsocketname() returns ('127.0.0.1', <port>)

    yield host, port, server, active_writers

    await close_server(server, active_writers)


def test_sync_connect_write_trans_and_disconnect(sync_echo_server):
    host, port = sync_echo_server
    dev = SocketDevice(hostname=host, port=port)

    dev.connect()
    assert dev.is_connected() is True

    # trans (send and get response that contains ETX)
    resp = dev.trans("HELLO")
    assert isinstance(resp, bytes)
    assert resp.endswith(SEPARATOR)
    assert b"RESP:HELLO" in resp

    # trans (send and get response that contains ETX)
    resp = dev.trans("HELLO:REPLY-IN-PARTS")
    assert isinstance(resp, bytes)
    assert resp.endswith(SEPARATOR)
    assert b"RESP:HELLO" in resp

    # write-only (then direct read via trans)
    dev.write("PING:NO-REPLY")
    time.sleep(0.1)
    assert b"RESP:ONE" in dev.trans("ONE")

    with pytest.raises(DeviceTimeoutError):
        dev.read()

    dev.disconnect()
    assert dev.is_connected() is False


def test_sync_connect_read_failures(sync_echo_server):
    host, port = sync_echo_server
    dev = SocketDevice(hostname=host, port=port, read_timeout=None)

    dev.connect()
    assert dev.is_connected() is True

    # the following line will hang forever, don't do this now
    # dev.read()

    dev.disconnect()

    dev = SocketDevice(hostname=host, port=port, read_timeout=1.0)

    dev.connect()
    assert dev.is_connected() is True


def test_sync_connect_invalid_parameters_raises():
    dev = SocketDevice(hostname="", port=1234)
    with pytest.raises(ValueError):
        dev.connect()

    dev = SocketDevice(hostname="127.0.0.1", port=0)
    with pytest.raises(ValueError):
        dev.connect()


def test_sync_connect_failures():
    dev = SocketDevice(hostname="127.0.0.1", port=1234)
    with pytest.raises(ConnectionError):
        dev.connect()

    # Connecting to a non-routable IP address results in a timeout error.
    # The port number doesn't really matter as long as it is not 0 (zero) because
    # that is checked by the SocketDevice constructor

    # dev = SocketDevice(hostname="10.255.255.1", port=25)
    # dev = SocketDevice(hostname="10.255.255.0", port=20)
    dev = SocketDevice(hostname="10.0.0.0", port=1)
    with pytest.raises(TimeoutError):
        dev.connect()

    # Connecting to a server where a port is blocked by the firewall also results in a timeout error
    dev = SocketDevice(hostname="www.google.com", port=81)
    with pytest.raises(TimeoutError):
        dev.connect()

    # Check proper handling of address name errors, " " is an invalid hostname
    dev = SocketDevice(hostname=" ", port=81)
    with pytest.raises(ConnectionError, match="Socket address info error"):
        dev.connect()


def test_sync_reconnect_behaviour(sync_echo_server):
    host, port = sync_echo_server
    dev = SocketDevice(hostname=host, port=port)

    dev.connect()
    assert dev.is_connected() is True

    # reconnect should first disconnect then connect again
    dev.reconnect()
    assert dev.is_connected() is True

    dev.disconnect()
    assert dev.is_connected() is False


@pytest.mark.asyncio
async def test_async_connect_write_trans_and_disconnect(async_echo_server):
    host, port, server, _active_writers = async_echo_server
    dev = AsyncSocketDevice(hostname=host, port=port, connect_timeout=1.0, read_timeout=1.0)

    await dev.connect()
    assert dev.is_connected() is True

    # trans should return bytes that end with ETX and contain the echoed payload
    got = await dev.trans("CMD-ASYNC")
    assert isinstance(got, bytes)
    assert got.endswith(SEPARATOR)
    assert b"RESP:CMD-ASYNC" in got

    # write-only then read via trans
    # since our test server doesn't interpret what is sent and always returns something,
    # you will also find 'W1' in the response of trans(), then need to do a read() again.
    await dev.write("W1:NO-REPLY")
    await asyncio.sleep(0.1)
    assert b"RESP:W2" in await dev.trans("W2")

    # Confirm there is not more data on the socket
    with pytest.raises(DeviceTimeoutError):
        await dev.read()

    # disconnect
    await dev.disconnect()
    assert dev.is_connected() is False


@pytest.mark.asyncio
async def test_async_read_timeout_raises_device_timeout(tmp_path):
    """
    Start a server that accepts data but intentionally does not send the ETX,
    so AsyncSocketDevice.read() should raise DeviceTimeoutError due to read_timeout.
    """

    # handler that reads but does NOT send ETX (simulate stuck/partial response)
    async def partial_handler(reader, writer):
        try:
            _ = await reader.read(4096)  # consume what client sends
            # intentionally do not write ETX back; keep connection open
            await asyncio.sleep(2.0)
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass

    server = await asyncio.start_server(partial_handler, "127.0.0.1", 0)
    host, port = server.sockets[0].getsockname()[:2]

    dev = AsyncSocketDevice(hostname=host, port=port, connect_timeout=1.0, read_timeout=0.2)
    try:
        await dev.connect()
        # write a command; the server won't return ETX so readuntil should time out
        with pytest.raises(DeviceTimeoutError):
            await dev.trans("NOETX")
    finally:
        await dev.disconnect()
        server.close()
        await server.wait_closed()


@pytest.mark.asyncio
async def test_async_cleanup_on_broken_connection(async_echo_server):
    host, port, server, active_writers = async_echo_server
    dev = AsyncSocketDevice(hostname=host, port=port, connect_timeout=1.0, read_timeout=1.0)

    # connect and then shut down server to simulate connection drop
    await dev.connect()
    assert dev.is_connected() is True

    logger.info("Before server close")

    # close server so subsequent reads/writes fail
    await close_server(server, active_writers)

    # reads/writes should raise DeviceConnectionError and device should clean up its state
    with pytest.raises(DeviceConnectionError):
        logger.info("Before read()")
        await dev.read()
        logger.info("After read()")

    with pytest.raises(DeviceConnectionError):
        await dev.write("X")

    assert dev.is_connected() is False



import asyncio
import sys
import textwrap
import threading
import time
from asyncio import subprocess
from pathlib import Path
from typing import Optional

import pytest

from egse.connect import AsyncServiceConnector
from egse.connect import BackoffStrategy
from egse.connect import ConnectionState
from egse.connect import JitterStrategy
from egse.connect import get_endpoint
from egse.device import DeviceConnectionError
from egse.log import logging
from egse.registry import is_service_registry_active
from egse.socketdevice import AsyncSocketDevice

logger = logging.getLogger("egse.test.connect")


@pytest.mark.asyncio
async def test_successful_connection_async():
    """Test that an AsyncServiceConnector establishes and reports a healthy connection.

    Defines a minimal AsyncServiceConnector subclass whose connect_to_service() and health_check() always succeed. A
    background coroutine repeatedly calls attempt_connection() and performs periodic health checks. After a short
    delay the test asserts that the connector is connected and that health_check() returns True, verifying normal
    connection and monitoring behavior.
    """

    from egse.connect import AsyncServiceConnector
    from egse.connect import ConnectionState

    class MyServiceConnector(AsyncServiceConnector):
        def __init__(self, service_name: str):
            super().__init__(service_name)

        async def connect_to_service(self) -> bool:
            return True

        async def health_check(self) -> bool:
            return True

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning(f"{connector.service_name} health check failed, marking as disconnected.")

            await asyncio.sleep(1)  # Check every second

    async def run_main_test(connector: AsyncServiceConnector):
        assert connector.is_connected()
        assert await connector.health_check()

    task_running = True

    connector = MyServiceConnector("my_service")

    task = asyncio.create_task(manage_my_service_connection(connector))

    await asyncio.sleep(1.0)

    await run_main_test(connector)

    task_running = False

    task.cancel("end-of-test-cancelling")
    try:
        await task
    except asyncio.CancelledError:
        logger.info("Task manage_my_service_connection was cancelled as expected.")


@pytest.mark.asyncio
async def test_unsuccessful_connection_async():
    """Test that an AsyncServiceConnector remains disconnected when connection attempts fail.

    Creates a minimal AsyncServiceConnector whose connect_to_service() always fails and whose health_check() reports
    unhealthy. A background coroutine repeatedly calls attempt_connection() and performs periodic health checks.
    After a short delay the test asserts the connector is not connected and that health_check() returns False,
    verifying failure handling and monitoring.
    """

    from egse.connect import AsyncServiceConnector
    from egse.connect import ConnectionState

    class MyServiceConnector(AsyncServiceConnector):
        def __init__(self, service_name: str):
            super().__init__(service_name)

        async def connect_to_service(self) -> bool:
            logger.warning(f"Couldn't connect to service {self.service_name}")
            return False

        async def health_check(self) -> bool:
            return False

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning(f"{connector.service_name} health check failed, marking as disconnected.")

            await asyncio.sleep(1)  # Check every second

    async def run_main_test(connector: AsyncServiceConnector):
        assert not connector.is_connected()
        assert not await connector.health_check()

    task_running = True

    connector = MyServiceConnector("my_service")

    task = asyncio.create_task(manage_my_service_connection(connector))

    await asyncio.sleep(1.0)

    await run_main_test(connector)

    task_running = False
    task.cancel("end-of-test-cancelling")
    try:
        await task
    except asyncio.CancelledError:
        logger.info("Task manage_my_service_connection was cancelled as expected.")


@pytest.mark.asyncio
async def test_retries_connection_async():
    """Test that an AsyncServiceConnector eventually connects after transient failures.

    Defines a connector that simulates a fixed number of failed connect_to_service() attempts
    before succeeding and exposing a connection object. A background task calls attempt_connection()
    repeatedly and runs health checks. The test asserts the initial disconnected state, waits
    long enough for retries to complete, then asserts the connector is connected, health_check()
    is True, and the returned connection object matches the expected value.
    """

    from egse.connect import AsyncServiceConnector
    from egse.connect import ConnectionState

    class MyServiceConnector(AsyncServiceConnector):
        def __init__(self, service_name: str, max_attempts: int = 3):
            super().__init__(service_name)
            self.attempts = 0
            self.max_attempts = max_attempts
            self.connection = None

        async def connect_to_service(self) -> bool:
            if self.attempts < self.max_attempts:
                logger.warning(f"Couldn't connect to service {self.service_name}")
                self.attempts += 1
            else:
                # self.attempts = 0
                logger.info(f"Connected to {self.service_name}.")
                self.connection = "I am a socket or transport object"

            return self.connection is not None

        async def disconnect_from_service(self) -> None:
            logger.info(f"Disconnecting from {self.service_name}..")
            self.connection = None
            self.state = ConnectionState.DISCONNECTED

        async def health_check(self) -> bool:
            if self.is_connected():
                return True
            else:
                return False

        def get_connection(self):
            return self.connection

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning("service health check failed, marking as disconnected")

            await asyncio.sleep(1.0)  # Check every second

    async def run_main_test(connector: AsyncServiceConnector):
        assert not connector.is_connected()
        assert not await connector.health_check()

        logger.info("Waiting for service to connect...")
        await asyncio.sleep(15.0)

        assert connector.is_connected()
        assert await connector.health_check()
        # FIXME: this method should be a method of the Connector base class, a connector
        #        should always be able to return its underlying connection object.
        assert connector.get_connection() == "I am a socket or transport object"

    task_running = True

    # Beware this test illustrates the exponential backoff + equal jitter strategies.
    # So, for 3 attempts, you should calculate a sleep() time of about 15s in the
    # run_main_test() before the service will be connected.

    connector = MyServiceConnector("my_service", max_attempts=3)

    task = asyncio.create_task(manage_my_service_connection(connector))

    await asyncio.sleep(1.0)

    await run_main_test(connector)

    task_running = False

    task.cancel("end-of-test-cancelling")
    try:
        await task
    except asyncio.CancelledError:
        logger.info("Task manage_my_service_connection was cancelled as expected.")

    await connector.disconnect_from_service()


@pytest.mark.asyncio
async def test_circuit_break_async():
    """Test that an AsyncServiceConnector ends up in a circuit break after # failures.

    Defines a connector that simulates a fixed number of failed connect_to_service() attempts
    before going into a circuit break. A background task calls attempt_connection()
    repeatedly and runs health checks. The test asserts the initial disconnected state, waits
    long enough for retries to complete, then asserts the connector still not connected, health_check()
    is False, and the state is in circuit break.
    """

    from egse.connect import AsyncServiceConnector
    from egse.connect import ConnectionState

    class MyServiceConnector(AsyncServiceConnector):
        def __init__(self, service_name: str):
            super().__init__(service_name)
            self.attempts = 0
            self.max_attempts = 6  # > max_failures_before_circuit_break
            self.connection = None

        async def connect_to_service(self) -> bool:
            if self.attempts < self.max_attempts:
                logger.warning(f"Couldn't connect to service {self.service_name}")
                self.attempts += 1
            else:
                # self.attempts = 0
                logger.info(f"Connected to {self.service_name}.")
                self.connection = "I am a socket or transport object"

            return self.connection is not None

        async def disconnect_from_service(self) -> None:
            logger.info(f"Disconnecting from {self.service_name}..")
            self.connection = None
            self.state = ConnectionState.DISCONNECTED

        async def health_check(self) -> bool:
            if self.is_connected():
                return True
            else:
                return False

        def get_connection(self):
            return self.connection

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning("service health check failed, marking as disconnected")
            else:
                logger.warning(f"Service {connector.service_name} is not connected")

            await asyncio.sleep(1.0)  # Check every second

    async def run_main_test(connector: MyServiceConnector):
        assert not connector.is_connected()
        assert not await connector.health_check()

        await asyncio.sleep(30.0)

        assert not connector.is_connected()
        assert not await connector.health_check()
        assert connector.state == ConnectionState.CIRCUIT_OPEN

        await asyncio.sleep(connector.circuit_break_duration)

        connector.max_attempts = 3

        await asyncio.sleep(15.0)

        assert connector.is_connected()
        assert await connector.health_check()

    task_running = True

    # Beware this test illustrates the exponential backoff + equal jitter strategies.
    # So, for 3 attempts, you should calculate a sleep() time of about 15s in the
    # run_main_test() before the service will be connected.

    connector = MyServiceConnector("my_service")

    task = asyncio.create_task(manage_my_service_connection(connector))

    await asyncio.sleep(1.0)

    await run_main_test(connector)

    task_running = False

    task.cancel("end-of-test-cancelling")
    try:
        await task
    except asyncio.CancelledError:
        logger.info("Task manage_my_service_connection was cancelled as expected.")

    await connector.disconnect_from_service()


@pytest.mark.asyncio
async def test_socket_connection_async():
    class AsyncSocketServiceConnector(AsyncServiceConnector):
        """
        Async connector that uses AsyncSocketDevice (no threads).
        """

        def __init__(self, service_name: str, hostname: str, port: int, connect_timeout: float = 3.0):
            super().__init__(service_name)
            self.hostname = hostname
            self.port = port
            self.connect_timeout = connect_timeout
            self.device: Optional[AsyncSocketDevice] = None

        async def connect_to_service(self) -> bool:
            try:
                self.device = AsyncSocketDevice(self.hostname, self.port, connect_timeout=self.connect_timeout)
                await self.device.connect()
                return self.device.is_connected()
            except Exception as exc:
                logger.warning(f"{self.service_name}: connect_to_service failed: {exc}")
                self.device = None
                return False

        async def disconnect_from_service(self) -> None:
            if self.device:
                await self.device.disconnect()
            self.state = ConnectionState.DISCONNECTED

        async def health_check(self) -> bool:
            if not self.device or not self.device.is_connected():
                device_name = self.device.device_name if self.device else "unknown"
                logger.warning(f"Device {device_name} not connected.")
                return False
            try:
                # lightweight ping; adjust protocol as needed
                resp = await asyncio.wait_for(self.device.trans("PING\x03"), timeout=1.0)
                return bool(resp)
            except (DeviceConnectionError, asyncio.TimeoutError, Exception):
                logger.debug(f"{self.service_name}: health_check failed, cleaning up")
                if self.device:
                    try:
                        await self.device.disconnect()
                    except Exception:
                        pass
                    self.device = None
                return False

        def get_device(self) -> Optional[AsyncSocketDevice]:
            return self.device

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning("service health check failed, marking as disconnected")

            await asyncio.sleep(1.0)  # Check every second

    async def run_main_test(connector: AsyncSocketServiceConnector):
        assert connector.is_connected()
        assert await connector.health_check()

        await asyncio.sleep(10.0)

        assert connector.is_connected()
        assert await connector.health_check()
        device: Optional[AsyncSocketDevice] = connector.get_device()
        assert device is not None
        assert device.is_connected()
        response = await device.trans("some_command: pars\x03")
        logger.info(f"Response received: {response}")
        assert response.decode().startswith("ACK")

    server = None
    task_running = True

    connector = AsyncSocketServiceConnector("my_simple_unix_server", "localhost", 5555)

    task = asyncio.create_task(manage_my_service_connection(connector))

    await asyncio.sleep(5.0)  # have a few failures to connect

    # Needs the simple_server running on localhost, 5555
    # Simulate that the server is coming alive only now.
    server = await simple_server()

    # wait for the manage_my_service_connection to connect
    while not connector.is_connected():
        await asyncio.sleep(0.1)

    try:
        await run_main_test(connector)
    finally:
        task_running = False
        task.cancel("end-of-test-cancelling")
        while not task.done():
            await asyncio.sleep(0.1)
        await connector.disconnect_from_service()
        if server:
            server.terminate()


def test_connection_sync(attempts: int = 2):
    """Test that a synchronous ServiceConnector retries in a background thread until connected.

    Implements a ServiceConnector whose connect_to_service() fails for the first few attempts
    and succeeds thereafter. A background thread repeatedly calls attempt_connection() with sleeps
    between tries. The test starts the thread and waits, asserting that the connector becomes
    connected within the test's waiting loop, demonstrating retry behavior in a threaded context.
    """
    from egse.connect import ServiceConnector

    class MyDeviceConnector(ServiceConnector):
        def connect_to_service(self) -> bool:
            logger.info(f"Attempting to connect to {self.service_name}...{self.failure_count=}")
            # Simulate a connection attempt (succeeds after 3 tries)
            if self.failure_count >= attempts:
                logger.info("Connection successful!")
                return True
            logger.info("Connection failed.")
            return False

    def background_connect(connector: ServiceConnector):
        logger.info("Establishing connection...")

        count = 0
        while not connector.is_connected():
            logger.info(f"In background waiting for connection....try {count}")
            connector.attempt_connection()
            time.sleep(0.5)
            count += 1

        logger.info("Connection established.")

    connector = MyDeviceConnector("my_device")

    thread = threading.Thread(target=background_connect, args=(connector,))
    thread.daemon = True
    thread.start()

    while not connector.is_connected():
        logger.info("Waiting for connection...")
        time.sleep(2.0)

    time.sleep(1.0)

    connector.disconnect_from_service()


async def single_connection_client(host="127.0.0.1", port=5555):
    # This coroutine connects to a simple_server that needs to be started separately from a terminal.
    # The simple server reside in this same folder in the test area.
    # This function served the purpose of demonstrating that there is only a single connection
    # created with the server.
    try:
        reader, writer = await asyncio.open_connection(host, port)

        print("client: connected once")

        try:
            for msg in ("PING\x03", "some_command: pars\x03", "QUIT\x03"):
                writer.write(msg.encode())
                await writer.drain()
                data = await asyncio.wait_for(reader.readuntil(separator=b"\x03"), timeout=2.0)
                print("client got:", data)
                await asyncio.sleep(0.2)
        finally:
            # We only need to close the writer since it owns the underlying transport (socket).
            # When we read after these calls, the reader will get an EOF.
            writer.close()
            await writer.wait_closed()
            print("client: connection closed")

    except ConnectionRefusedError as exc:
        print(f"ConnectionRefusedError: {exc}")


async def simple_server():
    cmd = Path(__file__).parent / "simple_server.py"
    server = await subprocess.create_subprocess_exec(sys.executable, cmd)
    await asyncio.sleep(0.2)  # give the server time to start up

    logger.info(f"simple_server process should be running: {server=}")

    return server


def test_get_endpoint():
    if not is_service_registry_active():
        pytest.xfail("This test expects the Registry Service to be running.")

    # This test also assumes the log server is running and registered to the service registry

    assert get_endpoint("LOG_CS").startswith("tcp://")
    assert get_endpoint("LOG_CS").endswith(":6106")

    with pytest.raises(RuntimeError, match="No service registered as"):
        assert get_endpoint(service_type="")
    with pytest.raises(RuntimeError, match="No service registered as"):
        assert get_endpoint(service_type=None)
    with pytest.raises(RuntimeError, match="No service registered as"):
        assert get_endpoint(hostname="localhost", port=0)

    assert get_endpoint("localhost", port=6106) == "tcp://localhost:6106"


@pytest.mark.asyncio
async def test_backoff_and_jitter():
    logger.warning("This test_backoff_and_jitter test will take about 45 seconds.")

    from egse.connect import AsyncServiceConnector
    from egse.connect import ConnectionState

    class MyServiceConnector(AsyncServiceConnector):
        def __init__(
            self,
            service_name: str,
            backoff_strategy: BackoffStrategy = BackoffStrategy.EXPONENTIAL,
            jitter_strategy: JitterStrategy = JitterStrategy.EQUAL,
        ):
            super().__init__(service_name, backoff_strategy, jitter_strategy=jitter_strategy)
            self.attempts = 0
            self.max_attempts = 3
            self.connection = None

        def __str__(self):
            return self.service_name

        async def connect_to_service(self) -> bool:
            if self.attempts < self.max_attempts:
                logger.warning(f"Couldn't connect to service {self.service_name}")
                self.attempts += 1
            else:
                # self.attempts = 0
                logger.info(f"Connected to {self.service_name}.")
                self.connection = "I am a socket or transport object"

            return self.connection is not None

        async def disconnect_from_service(self) -> None:
            logger.info(f"Disconnecting from {self.service_name}..")
            self.connection = None
            self.state = ConnectionState.DISCONNECTED

        async def health_check(self) -> bool:
            if self.is_connected():
                logger.debug("Service is found to be healthy.")
                return True
            else:
                logger.debug("Service is found to be unhealthy.")
                return False

        def get_connection(self):
            return self.connection

    async def manage_my_service_connection(connector):
        while task_running:
            await connector.attempt_connection()

            # Health check if connected
            if connector.is_connected():
                if not await connector.health_check():
                    connector.state = ConnectionState.DISCONNECTED
                    logger.warning("service health check failed, marking as disconnected")

            await asyncio.sleep(1.0)  # Check every second

    async def run_main_test(connector: AsyncServiceConnector):
        assert not connector.is_connected()
        assert not await connector.health_check()

        await asyncio.sleep(15.0)

        assert connector.is_connected()
        assert await connector.health_check()
        assert connector.get_connection() == "I am a socket or transport object"

    # Beware this test illustrates the exponential backoff + equal jitter strategies.
    # So, for 3 attempts, you should calculate a sleep() time of about 15s in the
    # run_main_test() before the service will be connected.

    for backoff, jitter in (
        (BackoffStrategy.EXPONENTIAL, JitterStrategy.EQUAL),
        (BackoffStrategy.LINEAR, JitterStrategy.NONE),
        (BackoffStrategy.FIXED, JitterStrategy.PERCENT_10),
    ):
        logger.info(f"Starting connector with backoff={backoff.name} and jitter={jitter.name}...")

        connector = MyServiceConnector("my_service", backoff_strategy=backoff, jitter_strategy=jitter)

        task_running = True

        task = asyncio.create_task(manage_my_service_connection(connector))

        await asyncio.sleep(1.0)

        await run_main_test(connector)

        task_running = False

        task.cancel("end-of-test-cancelling")
        try:
            await task
        except asyncio.CancelledError:
            logger.info(f"Task manage_my_service_connection for {connector} was cancelled.")

        await connector.disconnect_from_service()


if __name__ == "__main__":
    # You can run the individual tests with `py test_connect.py`
    # Add those tests that are missing when you need them.

    print(
        textwrap.dedent(
            """\
        1. single_connection_client (simple_server should be running)
        2. test_socket_connection_async
        3. test_circuit_break_async
        4. test_backoff_and_jitter
        5. test_get_endpoint
        6. test_connection_sync
        7. test_retries_connection_async
        """
        )
    )
    x = input("Select a number for the test you want to execute: ")

    try:
        match int(x.strip()):
            case 1:
                asyncio.run(single_connection_client())
            case 2:
                asyncio.run(test_socket_connection_async())
            case 3:
                asyncio.run(test_circuit_break_async())
            case 4:
                asyncio.run(test_backoff_and_jitter())
            case 5:
                test_get_endpoint()
            case 6:
                test_connection_sync()
            case 7:
                asyncio.run(test_retries_connection_async())

    except asyncio.CancelledError:
        pass



"""Command protocol for the Aim-TTi TGF4000."""

import logging
from pathlib import Path

from egse.arbitrary_wave_generator.aim_tti.tgf4000 import Tgf4000Interface, Tgf4000Simulator, Tgf4000Controller
from egse.arbitrary_wave_generator.aim_tti.tgf4000_cs import Tgf4000ControlServer
from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.device import DeviceConnectionState
from egse.hk import read_conversion_dict, convert_hk_names
from egse.protocol import DynamicCommandProtocol
from egse.settings import Settings
from egse.setup import SetupError
from egse.system import format_datetime
from egse.zmq_ser import bind_address

_HERE = Path(__file__).parent
DEVICE_SETTINGS = Settings.load(filename="tgf4000.yaml", location=_HERE)
LOGGER = logging.getLogger("egse.arbitrary_wave_generator.aim_tti.tgf4000")


class Tgf4000Command(ClientServerCommand):
    """Command class for the Aim-TTi TGF4000 Control Server."""

    pass


class Tgf4000Protocol(DynamicCommandProtocol):
    """Command protocol for the Aim-TTi TGF4000 Control Server."""

    def __init__(self, control_server: Tgf4000ControlServer, device_id: str, simulator: bool = False):
        """Initialisation of a Aim-TTi TGF4000 protocol.

        Args:
            control_server (ControlServer): Aim-TTi TGF4000 Control Server.
            device_id (str): Device identifier, as per (local) settings and setup.
            simulator (bool): Whether to use a simulator as the backend.
        """

        super().__init__(control_server)

        try:
            self.hk_conversion_table = read_conversion_dict(
                self.get_control_server().get_storage_mnemonic(), use_site=False
            )
        except SetupError:
            self.hk_conversion_table = None

        self.simulator = simulator

        if self.simulator:
            self.tgf4000: Tgf4000Interface = Tgf4000Simulator(device_id)
        else:
            self.tgf4000: Tgf4000Interface = Tgf4000Controller(device_id)

        try:
            self.tgf4000.connect()
        except ConnectionError:
            LOGGER.warning("Couldn't establish connection to the Aim-TTi TGF4000, check the log messages.")

    def get_bind_address(self) -> str:
        """Returns the bind address for the Aim-TTi TGF400 Control Server.

        Returns:
            Bind address for the Aim-TTi TGF400 Control Server.
        """

        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_device(self) -> Tgf4000Interface:
        """Returns the Aim-TTi TGF400 interface.

        Returns:
            Aim-TTi TGF400 interface.
        """

        return self.tgf4000

    def get_status(self) -> dict:
        """Returns the status information for the Aim-TTi TGF400 Control Server.

        Returns:
            Status information for the Aim-TTi TGF400 Control Server.
        """

        status = super().get_status()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return status

        # TODO Add device-specific status information

        return status

    def get_housekeeping(self) -> dict:
        """Returns the housekeeping information for the Aim-TTi TGF400 Control Server.

        Returns:
            Housekeeping information for the Aim-TTi TGF400 Control Server.
        """

        result = dict()
        result["timestamp"] = format_datetime()

        # TODO

        if self.hk_conversion_table:
            return convert_hk_names(result, self.hk_conversion_table)
        return result

    def is_device_connected(self) -> bool:
        """Checks whether the Aim-TTi TGF400 is connected.

        Returns:
            True if the Aim-TTi TGF400 is connected; False otherwise.
        """

        return self.tgf4000.is_connected()



import psutil
import time
from rich.live import Live
from rich.table import Table
from rich.console import Console
import threading


class ProcessDashboard:
    def __init__(self):
        self.console = Console()
        self.running = True

    def generate_table(self):
        table = Table(title="Live Process Monitor")

        table.add_column("PID", style="cyan", width=8)
        table.add_column("Name", style="magenta", width=15)
        table.add_column("CPU%", justify="right", style="yellow", width=8)
        table.add_column("Memory%", justify="right", style="blue", width=8)
        table.add_column("Status", style="green", width=10)
        table.add_column("Command", style="white")

        # Get top processes by CPU
        processes = []
        for proc in psutil.process_iter(["pid", "name", "cpu_percent", "memory_percent", "status", "cmdline"]):
            try:
                info = proc.info
                if info["cpu_percent"] is not None:
                    processes.append(info)
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

        # Sort by CPU usage
        processes.sort(key=lambda x: x.get("cpu_percent", 0), reverse=True)

        for proc in processes[:15]:  # Top 15
            cmdline = proc.get("cmdline", [])
            command = " ".join(cmdline) if cmdline else proc["name"]

            table.add_row(
                str(proc["pid"]),
                proc["name"][:14],
                f"{proc.get('cpu_percent', 0):.1f}",
                f"{proc.get('memory_percent', 0):.1f}",
                proc.get("status", "unknown"),
                command[:80] + "..." if len(command) > 40 else command,
            )

        return table

    def run(self):
        with Live(self.generate_table(), refresh_per_second=2) as live:
            while self.running:
                time.sleep(0.5)
                live.update(self.generate_table())

    def stop(self):
        self.running = False


# Usage
dashboard = ProcessDashboard()
try:
    dashboard.run()
except KeyboardInterrupt:
    dashboard.stop()



"""Quick test to understand version.py behavior"""


def test_with_block_scope():
    """Test that variables assigned in with blocks persist."""

    from pathlib import Path
    from egse.system import chdir

    # Test 1: Variable should persist after with block

    with chdir(Path.cwd()):
        test_var = "hello"

    print(f"test_var after with block: {test_var}")  # Should print "hello"

    # Test 2: What happens in get_version_installed?

    from egse.version import get_version_installed

    result = get_version_installed("nonexistent-package-xyz-123")
    print(f"get_version_installed result: {repr(result)}")

    # Test 3: What happens in get_version_from_git?

    from egse.version import get_version_from_git
    import tempfile

    with tempfile.TemporaryDirectory() as tmpdir:
        result = get_version_from_git(tmpdir)
        print(f"get_version_from_git result: {repr(result)}")


if __name__ == "__main__":
    test_with_block_scope()



from egse.log import logger
from egse.scpi import create_channel_list
from egse.tempcontrol.keithley.daq6510 import DEFAULT_BUFFER_1, DAQ6510Proxy
from egse.tempcontrol.keithley.daq6510_cs import is_daq6510_cs_active


def test_active():
    assert is_daq6510_cs_active() in (False, True)
    assert is_daq6510_cs_active(timeout=1.0) in (False, True)
    logger.info("DAQ6510 Control Server active: " + str(is_daq6510_cs_active()))


def test_proxy():
    with DAQ6510Proxy() as proxy:
        info = proxy.info()

        assert isinstance(info, str)

        idn = proxy.get_idn()
        assert isinstance(idn, str)

        assert info == idn

    logger.info("DAQ6510 Proxy test passed.")


def test_time():
    with DAQ6510Proxy() as proxy:
        dev_time_str = proxy.get_time()
        logger.info(f"Device time string: {dev_time_str}")
        logger.info("DAQ6510 Proxy time test passed.")


def test_measurement():
    """
    Test performing measurements with the DAQ6510 Proxy.
    1. Connect to the device using the proxy.
    2. Reset the device to ensure a clean state.
    3. Retrieve and log device information.
    4. Get and log buffer capacity and count.
    5. Configure sensors for temperature measurements on specified channels.
       - Channel 101: FRTD with PT100 in Celsius.
       - Channel 102: RTD with PT100 in Celsius.
    6. Setup measurements for the configured channels.
    7. Perform measurements and log the results.
    8. Finally, disconnect from the device.
    """
    count = 5
    interval = 1

    with DAQ6510Proxy() as daq:
        daq.reset()

        info = daq.info()
        assert isinstance(info, str)
        logger.info(f"Device info: {info}")

        buffer_capacity = daq.get_buffer_capacity()
        logger.info(f"buffer {DEFAULT_BUFFER_1} can still hold {buffer_capacity} readings")

        buffer_count = daq.get_buffer_count()
        logger.info(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")

        for sense, channels in [
            ({"TEMPERATURE": [("TRANSDUCER", "FRTD"), ("RTD:FOUR", "PT100"), ("UNIT", "CELSIUS")]}, "(@101)"),
            ({"TEMPERATURE": [("TRANSDUCER", "RTD"), ("RTD:TWO", "PT100"), ("UNIT", "CELSIUS")]}, "(@102)"),
        ]:
            daq.configure_sensors(channels, sense=sense)

        channels = create_channel_list((101, 102))
        logger.info(f"Channels: {channels}")

        daq.setup_measurements(channel_list=channels)

        meas_response = daq.perform_measurement(channel_list=channels, count=count, interval=interval)

        assert len(meas_response) == 2 * count, f"Expected {2 * count} measurements, got {len(meas_response)}"

        logger.info(f"Measurement response: {meas_response}")

        buffer_count = daq.get_buffer_count()
        logger.info(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")


if __name__ == "__main__":
    test_active()
    test_proxy()
    test_time()
    test_measurement()



"""
Script to test the DAQ6510 asynchronous monitor client functionality.

Run the script as follows:

    $ uv run py tests/script_daq6510_amon.py

This script connects to the DAQ6510 asynchronous monitor service, retrieves the current status,
starts polling for temperature readings on channels 101 and 102, and logs the readings.

The script assumes that the DAQ6510 is configured to read temperatures from the following channels:

- Channel 101: Temperature sensor 4-wire RTD
- Channel 102: Temperature sensor 2-wire RTD

The data will be stored in JSON format in the file 'temperature_readings.log'.

PREREQUISITES
-------------
Please ensure that the DAQ6510 asynchronous monitor service is running before executing this script.

    $ uv run py -m egse.tempcontrol.keithley.daq6510_amon

NOTE: The monitoring service directly connects to the DAQ6510 instrument, so make sure that no other
      process (e.g., DAQ6510 control server) is currently using the instrument.

"""

import json
import logging
import time

from egse.tempcontrol.keithley.daq6510_amon import DAQMonitorClient
from egse.tempcontrol.keithley.daq6510_amon import DAQ_MON_CMD_PORT

logger = logging.getLogger("egse.daq6510-client")


def daq6510_mon():
    with DAQMonitorClient(server_address="localhost", port=DAQ_MON_CMD_PORT) as client:
        # Get current status
        status = client.get_status()
        logger.info(f"Service status: {json.dumps(status, indent=4)}")

        # Set polling interval to 1s
        response = client.set_interval(1.0)
        logger.info(f"Set interval response: {response}")

        # Start polling with custom settings
        response = client.start_polling(channels=["101", "102"], interval=1.0)
        logger.info(f"Start polling response: {response}")

        # Wait for a while
        logger.info("Sleeping for 5.0s...")
        time.sleep(5.0)

        # doing some measurements in between ...
        response = client.get_reading(["101"])
        logger.info(f"Reading: {response = }")
        response = client.get_reading(["102"])
        logger.info(f"Reading: {response = }")

        # Wait for another 5.0s
        logger.info("Sleeping for 5.0s...")
        time.sleep(5.0)

        # another measurement during polling...
        response = client.get_reading(["101", "102"])
        logger.info(f"Reading: {response = }")

        # and get the last reading from the polling loop
        response = client.get_last_reading()
        logger.info(f"Reading: {response = }")

        # Wait for another 1.0s
        logger.info("Sleeping for 1.0s...")
        time.sleep(1.0)

        # Change polling interval
        response = client.set_interval(3.0)
        logger.info(f"Set interval response: {response}")

        # Wait for a while
        logger.info("Sleeping for 10.0s...")
        time.sleep(10.0)

        # Stop polling
        response = client.stop_polling()
        logger.info(f"Stop polling response: {response}")


if __name__ == "__main__":
    daq6510_mon()



import subprocess
import sys
import textwrap
import time

import pytest

from egse.dummy import is_dummy_dev_active
from egse.log import logger
from egse.process import SubProcess
from egse.process import is_process_running
from egse.system import Timer
from egse.system import waiting_for


def test_dummy_dev_1():
    print()

    if rc := is_process_running(["egse.dummy", "start-dev"]):
        pytest.xfail(f"dummy dev is{' ' if rc else ' not '}running...")

    dummy_dev_start = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["start-dev"])
    dummy_dev_start.execute()

    # We run a Timer here to log how long it takes to start the dummy device
    try:
        with Timer("dummy dev start"):
            waiting_for(is_dummy_dev_active, timeout=5.0)
    except TimeoutError:
        pytest.xfail("dummy dev should be active by now...")

    dummy_dev_stop = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["stop-dev"])
    dummy_dev_stop.execute()

    # We run a Timer here to log how long it takes to stop the dummy device
    try:
        with Timer("dummy dev stop"):
            waiting_for(lambda: not is_process_running(["egse.dummy", "start-dev"]), timeout=5.0)
    except TimeoutError:
        pytest.xfail("dummy dev should not be running anymore...")


def test_dummy_dev_2():
    print()

    rc = is_process_running(["egse.dummy", "start-dev"])
    logger.info(f"dummy dev is{' ' if rc else ' not '}running...")

    dummy_dev_start_1 = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["start-dev"])
    rc = dummy_dev_start_1.execute()
    logger.info(f"The dummy_dev_start_1 subprocess is{' ' if rc else ' not '}running...")

    logger.info(
        f"Status dummy dev 1: "
        f"{dummy_dev_start_1.returncode()=}, "
        f"{dummy_dev_start_1.exists()=}, "
        f"{dummy_dev_start_1.is_running()=}, "
        f"{dummy_dev_start_1.pid=}"
    )

    # We run a Timer here to log how long it takes to start the dummy device
    try:
        with Timer("dummy dev 1 start"):
            waiting_for(is_dummy_dev_active, timeout=5.0)
    except TimeoutError:
        pytest.xfail("dummy dev 1 should be active by now...")

    # Since the first dummy device is already running, this second start should fail

    dummy_dev_start_2 = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["start-dev"])
    rc = dummy_dev_start_2.execute()
    logger.info(f"The dummy_dev_start_2 subprocess {'started' if rc else 'not started'}.")

    # I cannot test here with is_dummy_dev_active() because the first process is already running
    time.sleep(2.0)  # About the time needed to start the process

    logger.info(
        f"Status dummy dev 2: "
        f"{dummy_dev_start_2.returncode()=}, "
        f"{dummy_dev_start_2.exists()=}, "
        f"{dummy_dev_start_2.is_running()=}, "
        f"{dummy_dev_start_2.pid=}"
    )

    # I cannot test here with is_process_running() because the first process is already running
    if dummy_dev_start_2.is_running():
        logger.info("dummy dev 2 is running...but shouldn't be running!")

    dummy_dev_stop = SubProcess("Dummy Device", [sys.executable, "-m", "egse.dummy"], ["stop-dev"])
    dummy_dev_stop.execute()

    # We run a Timer here to log how long it takes to stop the dummy device
    try:
        with Timer("dummy dev stop"):
            waiting_for(lambda: not is_process_running(["egse.dummy", "start-dev"]), timeout=5.0)
    except TimeoutError:
        pytest.xfail("dummy dev should not be running anymore...")


def test_dummy_dev_3():
    rc = is_process_running(["egse.dummy", "start-dev"])
    logger.info(f"dummy dev is{' ' if rc else ' not '}running...")

    dummy_dev = subprocess.Popen(
        [sys.executable, "-m", "egse.dummy", "start-dev"],
        stdout=None,
        stderr=None,
        stdin=subprocess.DEVNULL,
        close_fds=True,
    )

    logger.info(f"Status dummy dev 3: {dummy_dev.returncode=}, {dummy_dev.pid=}, {dummy_dev.args=}")

    time.sleep(1.0)

    dummy_dev.terminate()

    with Timer("dummy dev wait"):
        rc = dummy_dev.wait()

    rc = is_process_running(["egse.dummy", "start-dev"])
    logger.info(f"dummy dev is{' ' if rc else ' not '}running...")


if __name__ == "__main__":
    # You can run the individual tests with `py -m tests/test_dummy`
    # Add those tests that are missing when you need them.

    print(
        textwrap.dedent(
            """\
            1. start dummy device test 1 - start process, wait until active, stop process
            2. start dummy device test 2 - start process, try to start another, stop process
            3. start dummy device test 3 - start process with Popen, terminate it
            0. Exit
            """
        )
    )
    x = input("Select a number for the test you want to execute: ")

    try:
        match int(x.strip()):
            case 1:
                test_dummy_dev_1()
            case 2:
                test_dummy_dev_2()
            case 3:
                test_dummy_dev_3()
            case _:
                print("Exiting...")

    except Exception as exc:
        logger.error(f"Caught {type(exc).__name__}: {exc}")



from types import NoneType

from egse.setup import Setup
from egse.state import GlobalState


def test_state():
    print()
    print(f"{GlobalState.setup=}")

    assert isinstance(GlobalState.setup, (Setup, NoneType))
    assert isinstance(GlobalState.load_setup(), (Setup, NoneType))

    print(f"{GlobalState.setup=}")

    assert GlobalState.load_setup() is GlobalState.setup
    assert GlobalState.setup is not None

    print(f"{GlobalState.setup=}")
    print(f"{GlobalState.load_setup()=}")
    print(f"{GlobalState.setup=}")



from egse.mixin import DynamicCommandMixin
from egse.power_supply.kikusui import PROXY_TIMEOUT
from egse.power_supply.kikusui.pmx import PmxInterface
from egse.power_supply.kikusui.pmx_a import CS_SETTINGS
from egse.power_supply.kikusui.pmx_devif import PmxEthernetInterface
from egse.proxy import DynamicProxy
from egse.registry.client import RegistryClient
from egse.zmq_ser import connect_address


class PmxAInterface(PmxInterface):
    """Base class for KIKUSUI PMX-A power supply units."""

    def __init__(self, device_id: str):
        """Initialisation of a KIKUSUI PMX-A interface.

        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
        """

        super().__init__(device_id)


class PmxAController(PmxAInterface, DynamicCommandMixin):
    def __init__(self, device_id: str):
        """Initialisation of a controller for the KIKUSUI PMX-A device with the given identifier.

        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
        """

        super().__init__(device_id)

        self.transport = self.pmx_a = PmxEthernetInterface(device_id=device_id)

    # noinspection PyMethodMayBeStatic
    def is_simulator(self):
        return False

    def is_connected(self):
        """Checks whether the connection to the KIKUSUI PMX-A is open.

        Returns:
            True if the connection to the KIKUSUI PMX-A is open; False otherwise.
        """

        return self.transport.is_connected()

    def connect(self):
        """Opens the connection to the KIKUSUI PMX-A.

        Raises:
            PmxAError: When the connection could not be opened.
        """

        self.transport.connect()

    def disconnect(self):
        """Closes the connection to the KIKUSUI PMX-A.

        Raises:
            Dt8874Error: When the connection could not be closed.
        """

        self.transport.disconnect()

    def reconnect(self):
        """Re-connects to the KIKUSUI PMX-A."""

        self.transport.reconnect()


class PmxASimulator(PmxAInterface):
    def __init__(self, device_id: str):
        """Initialisation of a simulator for the KIKUSUI PMX-A device with the given identifier.

        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
        """

        super().__init__(device_id)

        self._is_connected = True

    # noinspection PyMethodMayBeStatic
    def is_simulator(self):
        return True

    # noinspection PyMethodMayBeStatic
    def is_connected(self):
        return self._is_connected

    def connect(self):
        self._is_connected = True

    def disconnect(self):
        self._is_connected = False

    def reconnect(self):
        self._is_connected = True


class PmxAProxy(DynamicProxy, PmxAInterface):
    def __init__(self, device_id: str):
        """Initialisation of proxy for the KIKUSUI PMX-A device with the given identifier.

        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
        """

        # super().__init__(device_id)

        hostname = CS_SETTINGS[device_id].get("HOSTNAME", "localhost")
        protocol = CS_SETTINGS[device_id].get("PROTOCOL", "tcp")
        commanding_port = CS_SETTINGS[device_id].get("COMMANDING_PORT", 0)
        service_type = CS_SETTINGS[device_id].get("SERVICE_TYPE", "pmx_a_cs")

        # Fixed ports -> Use information from settings

        if commanding_port != 0:
            super().__init__(connect_address(protocol, hostname, commanding_port))

        # Dynamic port allocation -> Use Registry Client

        else:
            with RegistryClient() as reg:
                service = reg.discover_service(service_type)

                if service:
                    protocol = service.get("protocol", "tcp")
                    hostname = service["host"]
                    port = service["port"]

                    super().__init__(connect_address(protocol, hostname, port), timeout=PROXY_TIMEOUT)

                else:
                    raise RuntimeError(f"No service registered as {service_type}")

        self.device_id = device_id



import nox

python_versions = ["3.9", "3.10", "3.11", "3.12", "3.13"]


@nox.session(python=python_versions)
def uv_tests(session: nox.Session):
    """
    Run the unit and regular tests using `uv`.
    """
    print(f"{session.python = }")

    py_version = ["--python", f"{session.python}"]

    uv_run_cmd = ["uv", "run", "--active", *py_version]
    uv_sync_cmd = ["uv", "sync", "--active", *py_version]

    # session.run_install("uv", "python", "pin", f"{session.python}")

    session.run_install("uv", "venv", *py_version)

    session.run_install(
        *uv_sync_cmd,  # "--all-packages",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
    )

    session.run(*uv_run_cmd, "python", "-V")
    session.run(*uv_run_cmd, "python", "-m", "site")
    session.run(*uv_run_cmd, "pytest", "-v", *session.posargs)


@nox.session(python=python_versions, default=False)
def py_tests(session: nox.Session):
    """
    Run the unit and regular tests using plain Python.
    """
    print(f"{session.python = }")

    session.install(".")
    session.install("pytest")
    session.install("pytest-cov")
    session.install("pytest-mock")

    session.run("python", "-V")
    session.run("python", "-m", "site")
    session.run("pytest", "-v", *session.posargs)



import textwrap

import rich

from egse.confman import ConfigurationManagerProxy
from egse.confman import is_configuration_manager_active
from egse.log import logger
from egse.response import Failure
from egse.setup import Setup


class CoreSetupProvider:
    def load_setup(self, setup_id: int = None, **kwargs) -> Setup | None:
        """
        This function loads the Setup corresponding with the given `setup_id`.

        Loading a Setup means:

        * that this Setup will also be loaded and activated in the configuration manager,
        * that this Setup will be available from the `GlobalState.setup`

        When no setup_id is provided, the current Setup is loaded from the configuration manager.

        When `from_disk` is True, the Setup will not be loaded from the configuration manager, but it
        will be loaded from disk. No interaction with the configuration manager happens in this case.

        Args:
            setup_id (int): the identifier for the Setup
            site_id (str): the name of the test house

        Returns:
            The requested Setup or None when the Setup could not be loaded from the \
            configuration manager.

        """
        logger.info(f"Loading Setup from core services, {setup_id=}, {kwargs=}")

        if not is_configuration_manager_active():
            logger.warning(
                textwrap.dedent(
                    """\
                    Could not reach the Configuration Manager to request the Setup.
    
                    Check if the Configuration Manager is running and why it can not be consulted. When it's
                    back on-line, do a 'load_setup()'.
                    """
                )
            )

            return None

        with ConfigurationManagerProxy() as proxy:
            setup = proxy.load_setup(setup_id) if setup_id else proxy.get_setup()
            if setup:
                logger.info(f"Loaded Setup {setup.get_id()}")
            return setup

    def submit_setup(self, setup: Setup, description: str, **kwargs) -> str | None:
        """
        Submit the given Setup to the Configuration Manager.

        When you submit a Setup, the Configuration Manager will save this Setup with the
        next (new) setup id and make this Setup the current Setup in the Configuration manager
        unless you have explicitly set `replace=False` in which case the current Setup will
        not be replaced with the new Setup.

        Args:
            setup (Setup): a (new) Setup to submit to the configuration manager
            description (str): one-liner to help identifying the Setup afterwards

        Returns:
            The Setup ID of the newly created Setup or None.
        """
        # We have not yet decided if this option should be made available. Therefore, we
        # leave it here as hardcoded True.

        # replace (bool): True if the current Setup in the configuration manager shall
        #                 be replaced by this new Setup. [default=True]
        replace: bool = True

        logger.info(f"Submitting Setup to core services, {setup=}, {description=}, {kwargs=}")

        if not is_configuration_manager_active():
            logger.warning(
                textwrap.dedent(
                    """\
                    Could not reach the Configuration Manager to submit the Setup.
    
                    Check if the Configuration Manager is running and why it can not be consulted. When it's
                    back on-line, redo the 'submit_setup()'.
                    """
                )
            )

            return None

        try:
            with ConfigurationManagerProxy() as proxy:
                setup = proxy.submit_setup(setup, description, replace)

            if setup is None:
                rich.print("[red]Submit failed for given Setup, no reason given.[/red]")
            elif isinstance(setup, Failure):
                rich.print(f"[red]Submit failed for given Setup[/red]: {setup}")
                setup = None
            elif replace:
                rich.print(
                    textwrap.dedent(
                        """\
                        [green]
                        Your new setup has been submitted and pushed to GitHub. The new setup is also
                        activated in the configuration manager. Load the new setup in your session with:
    
                            setup = load_setup()
                        [/]
                        """
                    )
                )
            else:
                rich.print(
                    textwrap.dedent(
                        f"""[dark_orange]
                        Your new setup has been submitted and pushed to GitHub, but has not been
                        activated in the configuration manager. To activate this setup, use the
                        following command:
    
                            setup = load_setup({str(setup.get_id())})
                        [/]
                        """
                    )
                )

            return setup.get_id() if setup is not None else None

        except ConnectionError:
            rich.print("Could not make a connection with the Configuration Manager, no Setup was submitted.")
        except NotImplementedError:
            rich.print(
                textwrap.dedent(
                    """\
                    Caught a NotImplementedError. That usually means the configuration manager is not running or
                    can not be reached. Check on the egse-server if the `cm_cs` process is running. If not you will
                    need to be restart the core services.
                    """
                )
            )

    def list_setups(**attr) -> None:
        logger.warning("Loading extension: list_setups() from cgse-core")

        return None

    def can_handle(self, source: str) -> bool:
        return source == "core-services"



import logging

from egse.tempcontrol.lakeshore.lakeshore336 import LakeShore336Simulator, SelfTestResult

MODULE_LOGGER = logging.getLogger(__name__)


def test_constructor():
    lakeshore = LakeShore336Simulator("TEST")
    assert lakeshore.device_id == "TEST"
    assert lakeshore.is_connected()


def test_disconnect():
    lakeshore = LakeShore336Simulator("TEST")
    assert lakeshore.is_connected()

    lakeshore.disconnect()
    assert not lakeshore.is_connected()


def test_reconnect():
    lakeshore = LakeShore336Simulator("TEST")

    lakeshore.reconnect()
    assert lakeshore.is_connected()


def test_info():
    lakeshore = LakeShore336Simulator("TEST")
    info = lakeshore.info()

    assert type(info) == tuple
    assert info[0] == "LakeShore"
    assert info[1] == "LSCI336"
    assert info[2] == "1234567/1234567"
    assert info[3] == 1.0


def test_get_temperature():
    with LakeShore336Simulator("TEST") as sim:
        assert -100.0 < sim.get_temperature("A") < 25.0


def test_get_pid_parameters():
    with LakeShore336Simulator("TEST") as sim:
        for output_channel in [1, 2]:
            sim.set_pid_parameters(output_channel, 1 * output_channel, 2 * output_channel, 3 * output_channel)
            assert sim.get_pid_parameters(output_channel) == (
                1 * output_channel,
                2 * output_channel,
                3 * output_channel,
            )


def test_get_heater_output():
    with LakeShore336Simulator("TEST") as sim:
        for output_channel in [1, 2]:
            sim.set_heater_output(output_channel, 3 * output_channel)
            assert sim.get_heater_output(output_channel) == 3 * output_channel


def test_get_selftest_result():
    with LakeShore336Simulator("TEST") as sim:
        assert sim.get_selftest_result() == SelfTestResult.NO_ERRORS_FOUND.value



import subprocess
from pathlib import Path
from typing import TextIO, Annotated

import sys
import rich
import typer

from egse.system import all_logging_disabled, redirect_output_to_log

tcu = typer.Typer(name="tcu", help="Ariel Telescope Control Unit (TCU)", no_args_is_help=True)


@tcu.command(name="start")
def start_tcu(
    simulator: Annotated[
        bool, typer.Option("--simulator", "--sim", help="use a device simulator as the backend")
    ] = False,
):
    """Start the Ariel TCU Control Server.

    The Control Server is always started in the background.
    """

    rich.print(f"Starting the Ariel TCU Control Server - {simulator = }")
    out = redirect_output_to_log("tcu_cs.start.log")

    cmd = [sys.executable, "-m", "egse.ariel.tcu.tcu_cs", "start"]
    if simulator:
        cmd.append("--simulator")

    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


@tcu.command(name="stop")
def stop_tcu():
    """Stops the Ariel TCU Control Server."""

    rich.print("Stopping the Ariel TCU Control Server")
    out = redirect_output_to_log("tcu_cs.stop.log")

    cmd = [sys.executable, "-m", "egse.ariel.tcu.tcu_cs", "stop"]
    subprocess.Popen(cmd, stdout=out, stderr=out, stdin=subprocess.DEVNULL, close_fds=True)


@tcu.command(name="status")
def status_tcu():
    """Prints the status information about the Ariel TCU Control Server."""

    with all_logging_disabled():
        from egse.ariel.tcu import tcu_cs

        tcu_cs.status()


if __name__ == "__main__":
    tcu()



from unittest.mock import Mock

mock = Mock()
mock


def do_something(json_lib):
    data = {"key": "value"}
    json_str = json_lib.dumps(data)
    print(f"Serialized JSON: {json_str}")
    parsed_data = json_lib.loads(json_str)
    print(f"Parsed Data: {parsed_data}")


# Pass mock as an argument to do_something()
do_something(mock)


# Patch the json library
json = mock


# Access attributes and call methods on the mocked json library
json.some_attribute


# Call a method on the mocked json library
json.do_something()


json = Mock()
json.dumps()

from unittest.mock import Mock

json = Mock()
json.loads('{"key": "value"}')

json.loads.assert_called()
json.loads.assert_called_once()
json.loads.assert_called_with('{"key": "value"}')
json.loads.assert_called_once_with('{"key": "value"}')

json.loads()
json.loads.assert_called_once()


json = Mock()

json.loads(s='{"key": "value"}')


json.loads.assert_called_with('{"key": "value"}')


json.loads.assert_called_with(s='{"key": "value"}')


from unittest.mock import Mock

json = Mock()
json.loads('{"key": "value"}')


# Number of times you called loads():
json.loads.call_count


# The last loads() call:
json.loads.call_args


# List of loads() calls:
json.loads.call_args_list


# List of calls to json's methods (recursively):
json.method_calls

json.loads.assert_called_once()
json.loads.assert_called_once_with('{"key": "value"}')
json.loads.assert_called_with(**{"key": "value"})  # This will fail



import inspect
import multiprocessing
import threading
import types
from pathlib import Path
from typing import Union, Any, Callable

import sys
import typer
from PyQt5.QtCore import QLockFile
from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QTabWidget,
    QWidget,
    QVBoxLayout,
    QComboBox,
    QLineEdit,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QMessageBox,
)

from egse.ariel.tcu import (
    TcuMode,
    NUM_M2MD_AXES,
    NUM_M2MD_POSITIONS,
    NUM_TSM_PROBES_PER_FRAME,
    NUM_TSM_FRAMES,
    AXIS_VELOCITY,
)
from egse.ariel.tcu.tcu import TcuInterface, TcuHex, TcuProxy
from egse.ariel.tcu.tcu_cs import is_tcu_cs_active
from egse.gui import QHLine
from egse.log import logging
from egse.observer import Observable, Observer
from egse.process import ProcessStatus
from egse.resource import get_resource
from egse.response import Failure
from egse.settings import Settings
from egse.system import do_every

MODULE_LOGGER = logging.getLogger(__name__)
TCU_COMPONENTS = ["GENERAL", "M2MD", "TSM", "HK"]
CTRL_SETTINGS = Settings.load("Ariel TCU Controller")

app = typer.Typer()


class TcuUIModel:
    """Model in the MVC pattern that makes the TCU UI application."""

    def __init__(self):
        """Initialisation of the TCU UI Model."""

        # This is used to generate the hex string that is sent to the Arduino, so it can be shown in the TCU UI
        # (Might be discontinued in the future)

        self.tcu_hex = TcuHex()

        # This is used to establish a connection to the TCU and actually send commands to it

        try:
            self.tcu_proxy: TcuProxy[TcuHex, None] = TcuProxy()
        except RuntimeError:
            MODULE_LOGGER.error("Could not connect to Ariel TCU Control Server")
            self.tcu_proxy: Union[TcuProxy, None] = None

    @staticmethod
    def build_dyn_cmds_list() -> dict:
        """Keep track of which dynamic commands have been defined.

        Per component of the TCU, we store a dictionary of the dynamic commands that have been defined for that
        component.  The keys in these dictionaries are the names of the dynamic commands, and the values are the
        functions that implement them.

        Returns:
            Dictionary that stores - per component of the TCU - a dictionary of the dynamic commands that have been
            defined for that component.
        """

        dyn_cmds = {}
        for component in TCU_COMPONENTS:
            dyn_cmds[component] = {}

        for cmd_name, func in inspect.getmembers(TcuInterface, inspect.isfunction):
            if hasattr(func, "target_comp"):
                target_comp = getattr(func, "target_comp")
                dyn_cmds[target_comp][cmd_name] = func

        return dyn_cmds


class TcuUiView(QMainWindow, Observable):
    """View in the MVC pattern that makes the TCU UI application."""

    def __init__(self):
        """Initialisation of the TCU UI View."""

        super(TcuUiView, self).__init__()
        Observable.__init__(self)

        self.setGeometry(300, 300, 500, 300)
        self.setWindowTitle("Telescope Control Unit")

        # Central widget = tabs per component

        self.tabs = QTabWidget()
        self.setCentralWidget(self.tabs)

    def build_tabs(self, dyn_cmds: dict, observer):
        """Build the tabs of the TCU UI view.

        For each component of the TCU commanding, a tab is created.  In this tab, you can find:

            - Drop-down menu with the list of available dynamic commands for the component,
            - If applicable: entry for cargo1,
            - If applicable: entry for cargo2,
            - "Run" button,
            - Command string (read-only),
            - Response (read-only).

        Args:
            dyn_cmds (dict): Dictionary that stores - per component of the TCU - a dictionary of the dynamic commands
                             that have been defined for that component.
            observer: Observer that will be notified when a command is run (i.e. when the "Run" button is clicked).
        """

        for tab_name, dyn_cmd_for_tab in dyn_cmds.items():
            tab = TabWidget(dyn_cmd_for_tab)
            tab.add_observer(observer)

            self.tabs.addTab(tab, tab_name)

        # When all tabs have been populated, select the first dynamic command in the drop-down menu in the first tab,
        # to make sure that the correct input parameters are requested (if applicable)

        self.tabs.currentChanged.connect(self.on_tab_changed)
        self.on_tab_changed(0)

    def on_tab_changed(self, _):
        """Takes action when another tab is selected.

        When a new tab has been selected, select the current dynamic command in the drop-down menu in the new tab,
        to make sure that the correct input parameters are requested (if applicable).
        """

        self.tabs.currentWidget().dyn_cmd_drop_down_menu_changed(
            self.tabs.currentWidget().dyn_cmd_drop_down_menu.currentText()
        )


class TabWidget(QWidget, Observable):
    """Tab in the TCU UI View."""

    def __init__(self, dyn_cmds: dict):
        """Initialisation of a tab in the TCU UI View.

        In this tab, you can find:

            - Drop-down menu with the list of available dynamic commands for the component,
            - If applicable: entry for cargo1,
            - If applicable: entry for cargo2,
            - "Run" button,
            - Command string (read-only),
            - Response (read-only).

        Args:
            dyn_cmds (dict): Dictionary that stores for one TCU component a dictionary of the dynamic commands
                             that have been defined for that component.
        """

        super().__init__()
        layout = QVBoxLayout(self)

        # Dynamic commands that are available for this component
        # (i.e. the dictionary of dynamic commands that have been defined for this component)

        self.dyn_cmds = dyn_cmds

        # Entries for cargo1 and/or cargo2 (if applicable)
        # We have to define these before we create the drop-down menu with the dynamic commands

        self.cargo1_layout = CargoLayout(1)
        self.cargo2_layout = CargoLayout(2)

        # Drop-down menu with the list of dynamic commands that are available for this component

        self.dyn_cmd_drop_down_menu = QComboBox()

        for cmd_name in self.dyn_cmds:
            self.dyn_cmd_drop_down_menu.addItem(cmd_name)
        self.dyn_cmd_drop_down_menu.currentTextChanged.connect(self.dyn_cmd_drop_down_menu_changed)

        # Button to send the command to the Arduino

        button = QPushButton("Run")
        button.clicked.connect(self.button_clicked)

        # Display the hex string tha is sent to the Arduino

        cmd_string_layout = QHBoxLayout()
        self.cmd_string = QLineEdit()
        self.cmd_string.setReadOnly(True)
        cmd_string_layout.addWidget(QLabel("Command string"))
        cmd_string_layout.addWidget(self.cmd_string)

        # Display the response that is received from the Arduino
        # (after the command has been sent by clicking the "Run" button)

        response_layout = QHBoxLayout()
        self.response = QLineEdit()
        self.response.setReadOnly(True)
        response_layout.addWidget(QLabel("Response"))
        response_layout.addWidget(self.response)

        # Assembly of all components in the tab

        layout.addWidget(self.dyn_cmd_drop_down_menu)  # Drop-down menu with available dynamic commands
        layout.addLayout(self.cargo1_layout)  # Cargo1
        layout.addLayout(self.cargo2_layout)  # Cargo2
        layout.addWidget(button)  # "Run" button
        layout.addWidget(QHLine())  # Horizontal separator
        layout.addLayout(cmd_string_layout)  # Hex string that is sent to the Arduino
        layout.addWidget(QHLine())  # Horizontal separator
        layout.addLayout(response_layout)  # Response that is received from the Arduino

    def dyn_cmd_drop_down_menu_changed(self, dyn_cmd_name):
        """Takes action when another dynamic command has been selected in the drop-down menu.

        Args:
            dyn_cmd_name (str): Name of the dynamic command that has been selected in the drop-down menu.
        """

        # Extract the signature of the selected function (1st item is always `self`)

        signature = inspect.signature(self.dyn_cmds[dyn_cmd_name])
        parameters = signature.parameters

        # `self` + cargo1 + cargo2 (the latter two may have a different name -> extract from signature)

        if len(parameters) == 3:
            self.cargo1_layout.set_visible(True)
            self.cargo1_layout.update_entry(list(parameters.keys())[-2])
            self.cargo2_layout.set_visible(True)
            self.cargo2_layout.update_entry(list(parameters.keys())[-1])

        # `self` + cargo2 (the latter may have a different name -> extract from signature)

        elif len(parameters) == 2:
            self.cargo1_layout.set_visible(False)
            self.cargo2_layout.set_visible(True)
            self.cargo2_layout.update_entry(list(parameters.keys())[-1])

        # `self` (no cargo)

        else:
            self.cargo1_layout.set_visible(False)
            self.cargo2_layout.set_visible(False)

        self.repaint()

    def button_clicked(self):
        """Takes action when the "Run" button is clicked.

        When the "Run" button is clicked, the observer (i.c. TCU UI Controller) is notified.  It will receive a tuple
        with the following information:

            - The tab widget in which the "Run" button has been clicked (needed to be able to fill out the response),
            - The dynamic command that has been selected in the drop-down menu (function),
            - The cargo1 value (if applicable),
            - The cargo2 value (if applicable).

        The TCU UI Controller will make sure that the hex string that is sent to the Arduino and the response that is
        received from the Arduino are displayed in the UI.  It will receive this information from the TCU UI Model.
        """

        self.notify_observers(
            (
                self,  # Tab widget in which the "Run" button has been clicked
                self.dyn_cmds[self.dyn_cmd_drop_down_menu.currentText()],  # Selected dynamic command (as a function)
                self.cargo1_layout.entry.text(),  # Cargo1 entry
                self.cargo2_layout.entry.text(),  # Cargo2 entry
            )
        )


class CargoLayout(QHBoxLayout):
    """Layout for the input argument(s) of the dynamic commands."""

    def __init__(self, cargo: int):
        """Initialisation of a layout for the given cargo argument (1 or 2).

        Args:
            cargo (int): Number of the cargo argument (1 or 2).
        """

        super().__init__()

        # Label for the argument ("cargo1" / "cargo2", depending on the input, as default)

        self.label = QLabel(f"cargo{cargo}")

        # By default, the entry is a QLineEdit

        self.entry = QLineEdit()

        self.addWidget(self.label)
        self.addWidget(self.entry)

    def update_entry(self, cargo_name: str):
        """Update the entry for the input argument with the given name.

        Updating the entry entails:

            - Updating the label with the name of the input argument,
            - Replacing the entry with the appropriate widget (this can be a line edit, a combo box, etc.).

        Args:
            cargo_name (str): Name of the input argument.
        """

        # Update the name of the input argument

        self.label.setText(cargo_name)

        # The `axis` argument is only used for M2MD commands, to denote the axis that should be commanded
        # -> Use a drop-down menu instead of the default QLineEdit

        if cargo_name == "axis":
            self.replace_entry_with_axis()

        # The `tcu_mode` argument is only used to change the TCU mode
        # -> Use a drop-down menu instead of the default QLineEdit

        elif cargo_name == "tcu_mode":
            self.replace_with_tcu_mode()

        # The `position` argument is only used for `sw_rs_xx_sw_rise` and `sw_rs_xx_sw_fall`, to denote the position
        # that should be commanded
        # -> Use a drop-down menu instead of the default QLineEdit (read the number of positions from the settings)

        elif cargo_name == "position":
            self.replace_with_int_combo_box(1, NUM_M2MD_POSITIONS)

        # The `probe` argument is only used for `tsm_adc_value_xx_currentn`, `tsm_adc_value_xx_biasn`,
        # `tsm_adc_value_xx_currentp`, and `tsm_adc_value_xx_biasp`
        # -> Use a drop-down menu instead of the default QLineEdit (read the number of probes from the settings)

        elif cargo_name == "probe":
            self.replace_with_int_combo_box(1, NUM_TSM_FRAMES * NUM_TSM_PROBES_PER_FRAME)

        elif cargo_name == "speed":
            self.replace_with_axis_speed_combo_box()

        # By default, the entry is a QLineEdit

        else:
            self.replace_with_line_edit()

    def replace_with_line_edit(self):
        """Replaces the entry with a QLineEdit (if it's not a QLineEdit already)."""

        if not isinstance(self.entry, QLineEdit):
            # Remove the current entry

            self.removeWidget(self.entry)
            self.entry.hide()
            self.entry.deleteLater()

            # Replace it with a QLineEdit

            self.entry = QLineEdit()
            self.insertWidget(1, self.entry)

    def replace_entry_with_axis(self):
        """Replaces the entry with an AxisComboBox (if it's not an AxisComboBox already)."""

        if not isinstance(self.entry, AxisComboBox):
            # Remove the current entry

            self.removeWidget(self.entry)
            self.entry.hide()
            self.entry.deleteLater()

            # Replace it with an AxisComboBox

            self.entry = AxisComboBox()
            self.insertWidget(1, self.entry)

    def replace_with_tcu_mode(self):
        """Replaces the entry with a TcuModeComboBox (if it's not a TcuModeComboBox already)."""

        if not isinstance(self.entry, TcuModeComboBox):
            # Remove the current entry

            self.removeWidget(self.entry)
            self.entry.hide()
            self.entry.deleteLater()

            # Replace it with a TcuModeComboBox

            self.entry = TcuModeComboBox()
            self.insertWidget(1, self.entry)

    def replace_with_int_combo_box(self, min_value: int, max_value: int):
        """Replaces the entry with an IntComboBox (if it's not an IntComboBox already).

        Args:
            min_value (int): Minimum value of the IntComboBox.
            max_value (int): Maximum value of the IntComboBox.
        """

        # Remove the current entry

        self.removeWidget(self.entry)
        self.entry.hide()
        self.entry.deleteLater()

        # Replace it with an IntComboBox

        self.entry = IntComboBox(min_value, max_value)
        self.insertWidget(1, self.entry)

    def replace_with_axis_speed_combo_box(self):
        """Replaces the entry with an AxisSpeedComboBox (if it's not an AxisSpeedComboBox already)."""

        if not isinstance(self.entry, AxisSpeedComboBox):
            # Remove the current entry

            self.removeWidget(self.entry)
            self.entry.hide()
            self.entry.deleteLater()

            # Replace it with an AxisSpeedComboBox

            self.entry = AxisSpeedComboBox()
            self.insertWidget(1, self.entry)

    def set_visible(self, visible: bool):
        """Changes the visibility of the label and the entry.

        Args:
            visible (bool): Indicates whether the label and the entry should be visible or not.
        """

        self.label.setVisible(visible)
        self.entry.setVisible(visible)


class AxisComboBox(QComboBox):
    def __init__(self):
        """Initialisation of a drop-down menu for the M2MD axes."""

        super().__init__()

        for axis in range(1, NUM_M2MD_AXES + 1):
            self.addItem(f"M2MD axis {axis}")
        self.setEditable(False)

    def text(self):
        """Converts the selected text to a valid input argument for the dynamic command."""

        return self.currentText()[-1]


class TcuModeComboBox(QComboBox):
    def __init__(self):
        """Initialisation of a drop-down menu for the TCU mode."""

        super().__init__()

        for tcu_mode in TcuMode:
            self.addItem(tcu_mode.name)
        self.setEditable(False)

    def text(self):
        """Converts the selected text to a valid input argument for the dynamic command."""

        return TcuMode[self.currentText()].value


class IntComboBox(QComboBox):
    def __init__(self, min_value: int, max_value: int):
        """Initialisation of a drop-down menu for a range of integers.

        Args:
            min_value (int): Minimum value in the drop-down menu.
            max_value (int): Maximum value in the drop-down menu.
        """

        super().__init__()

        self.addItems([str(x) for x in range(min_value, max_value + 1)])
        self.setEditable(False)

    def text(self):
        """Converts the selected text to a valid input argument for the dynamic command."""

        return int(self.currentText())


class AxisSpeedComboBox(QComboBox):
    def __init__(self):
        """Initialisation of a drop-down menu for the axis speed."""

        super().__init__()

        for speed in AXIS_VELOCITY:
            self.addItem(f"{speed}Hz")

        self.setEditable(False)

    def text(self):
        """Converts the selected text to a valid input argument for the dynamic command.

        The selected text is of the form "XHz", where X is the axis speed in Hz.  This has to be converted into a hex
        string that represents this axis speed.

        Returns:
            Axis speed in hex string format.
        """

        return AXIS_VELOCITY[int(self.currentText()[:-2])]


class TcuUiController(Observer):
    """Controller in the MVC pattern that makes the TCU UI application."""

    def __init__(self, model: TcuUIModel, view: TcuUiView):
        """Initialisation of the TCU UI Controller."""

        super().__init__()

        self.model = model
        self.view = view

        self.view.build_tabs(self.model.build_dyn_cmds_list(), self)

    def do(self, actions):
        # Abstract method from the Observer class, which we do not need here
        pass

    def update(self, changed_object):
        """Updates the TCU UI when the "Run" button is clicked in one of the tabs.

        In this context, updating means:

            - Generating the command string that will be sent to the Arduino and display it in the TCU UI View,
            - Sending the command to the Arduino and display the response in the TCU UI View.
        """

        origin_tab, func, cargo1, cargo2 = changed_object

        signature = inspect.signature(func)
        args = ()
        kwargs = {}

        parameters = signature.parameters

        if len(parameters) == 3:
            cargo1_name = list(parameters.keys())[-2]
            kwargs[cargo1_name] = cargo1

        if len(parameters) >= 2:
            cargo2_name = list(parameters.keys())[-1]
            kwargs[cargo2_name] = cargo2

        # Generate the command string + display

        cmd_string = call_unbound(func, self.model.tcu_hex, *args, **kwargs)
        origin_tab.cmd_string.setText(cmd_string.decode())

        # Execute the command + display the response

        if self.model.tcu_proxy:
            response = call_unbound(func, self.model.tcu_proxy, *args, **kwargs)

            if isinstance(response, Failure):
                origin_tab.response.setStyleSheet("background-color: red;")
                origin_tab.response.setText(str(response.cause))

            else:
                if hasattr(response, "decode") and callable(getattr(response, "decode")):
                    response = response.decode()

                origin_tab.response.setText(str(response))

        elif is_tcu_cs_active():
            self.model.tcu_proxy = TcuProxy()


def call_unbound(func: Callable, instance: object, *args: Any, **kwargs: Any) -> Any:
    """Executes the given function on the given instance with the given arguments.

    This helper handles three cases:

    1. `func` is already a bound method (has `__self__`) -> Call it directly.
    2. Prefer looking up the attribute on `instance` with `getattr(instance, func.__name__)`.
       This lets Python perform normal Method Resolution Order (MRO) so overrides on the instance's class or proxy
       wrappers are returned as bound methods.
    3. If look-up fails (no such attribute on the instance), fall back to binding the original function to `instance`
       using `types.MethodType` (this may call the base/class implementation).

    Args:
        func (Callable): Function (dynamic command) to execute.
        instance (object): Instance on which to execute the function.
        *args: Optional positional arguments for the function.
        **kwargs: Optional keyword arguments for the function.

    Returns:
        Any: Result of the function execution.
    """

    # Case 1: already bound -> call directly

    if getattr(func, "__self__", None) is not None:
        return func(*args, **kwargs)

    # Case 2: preferred lookup on the instance so Python performs normal MRO (Method Resolution Order) and returns
    # the appropriate bound method (honours overrides / proxy behaviour).

    try:
        bound = getattr(instance, func.__name__)
        return bound(*args, **kwargs)
    except AttributeError:
        # Case 3: Fall-back to binding the original function to the instance. This will create a bound method that
        # directly calls the function as if it were defined on the instance.

        bound = types.MethodType(func, instance)
        return bound(*args, **kwargs)


# def call_method_hex(instance: TcuHex, func: Callable, *args, **kwargs) -> Any:
#     """Executes the given dynamic command for the given TCU interface.
#
#     Args:
#         instance (TcuHex): TCU interface for which to execute the given dynamic command.
#         func (Callable): Dynamic command to execute.
#         *args: Optional positional arguments for the dynamic command.
#         **kwargs: Optional keyword arguments for the dynamic command.
#
#     Returns:
#        Response received from the TCU interface.
#     """
#
#     wrapper = instance.handle_dynamic_command(func)
#     result = wrapper(*args, **kwargs)
#
#     return result


@app.command()
def main():
    multiprocessing.current_process().name = "tcu_ui"
    lock_file = QLockFile(str(Path("~/tcu_ui.app.lock").expanduser()))

    tcu_app = QApplication(["-stylesheet", str(get_resource(":/styles/default.qss"))])
    # app_logo = get_resource(":/icons/logo-tcu.svg")
    # app.setWindowIcon(QIcon(str(app_logo)))

    if lock_file.tryLock(100):
        process_status = ProcessStatus()

        timer_thread = threading.Thread(target=do_every, args=(10, process_status.update))
        timer_thread.daemon = True
        timer_thread.start()

        # Create the TCU UI, following the MVC-model

        model = TcuUIModel()
        view = TcuUiView()
        controller = TcuUiController(model, view)
        view.add_observer(controller)

        view.show()

        return tcu_app.exec_()
    else:
        error_message = QMessageBox()
        error_message.setIcon(QMessageBox.Warning)
        error_message.setWindowTitle("Error")
        error_message.setText("The TCU GUI application is already running!")
        error_message.setStandardButtons(QMessageBox.Ok)

        return error_message.exec()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format=Settings.LOG_FORMAT_FULL)

    sys.exit(app())
